#!/usr/local/bin/perl
#
#  file: big.pl
#  auth: brad burdick (bburdick@uucom.com)
#  desc: Blocks Information Groper (BIG) implementation.
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package BIG;

use strict;
use Carp;

# version of perl we expect
require 5.004;


use vars qw($VERSION);

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/examples/client/big.pl#12 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# command line option processing
use Getopt::Long;


# some useful constants
use constant BXXP_PORT	=> 10288;


# where we find local libraries, if any
use lib "../../lib";


# Blocks client API
use Space::Mixer;

# spacekit log support
use Space::Log;

# spacekit XML support
use Space::XML qw(xml_escape_av);


# global variables

use vars qw($debug);

# debug level
$debug = 0;

# app name
my $prog;
($prog = $0) =~ s|^.*/||;


# default client options
my (%options) = (
	-debug      => "none",
	-logfile    => "logs/big.log",
	-mechanism  => "none",
	-passphrase => "",
	-passback   => undef,
	-port       => BXXP_PORT,
	-privacy    => "none",
	-server     => "sqa.invisible.net",
	-trace      => "",
	-username   => "",
);


# available command list and whether it is currently supported
my (%cmds) = (
	bulk   => 0,
	fetch  => 1,
	search => 1,
	store  => 1,
);

# get log package object
my $log;

# get mixer package object
my $mixer;


#
# main
#


#
# process cmd line options, if any
#
my (%useropts) = (
	debug      => \$options{-debug},
	logfile    => \$options{-logfile},
	mechanism  => \$options{-mechanism},
	passphrase => \$options{-passphrase},
	port       => \$options{-port},
	privacy    => \$options{-privacy},
	server     => \$options{-server},
	trace      => \$options{-trace},
	username   => \$options{-username},
);

my $version = 0;

my $result = GetOptions(\%useropts, "debug=s", "generic=s", "logfile=s",
                          "mechanism=s", "passphrase=s", "port=i", "privacy=s",
                          "server=s", "trace=s", "username=s",
                          "version" => \$version);

# display version info and exit
if ($version) {
	print version(), "\n";
	exit 0;
}


if (!$result || ($#ARGV < 0)) {
	die usage();
}


STDOUT->autoflush(1);

# valid command?
my $cmd;
foreach my $c (keys %cmds) {
	if ($ARGV[0] eq $c) {
		$cmd = shift(@ARGV);
		last;
	}
}

unless ($cmd) {
	adios(1, "user", "unknown command: $ARGV[0]");
}


# log token
my $ident = (split(/\./, $prog))[0];

$log = Space::Log->new($options{-logfile}, $ident);
unless ($log) {
	adios(1, "user", "log::new failed");
}

delete $options{-logfile};

my $server = $options{-server};
delete $options{-server};

$options{-passback} = \&do_passback;

if ($cmd eq "bulk") {
	$options{-bulk} = "true";
}


$log->entry("info", "begin $cmd");


$mixer = Space::Mixer->new();

my $response;
if ($mixer) {
	$response = $mixer->init($log, $server, \%options);
	if ($! == 0) {
		# NO-OP
		;
	}
	elsif ($! < 0) {
		$log->entry("user", "mixer::init", 
		            "$response->{code}: $response->{diagnostic}");

		undef $mixer;

		adios(1, "user", "$response->{code}: $response->{diagnostic}");
	}
	else {
		$log->entry("error", "mixer::init", 
		            "$response->{code}: $response->{diagnostic}");

		undef $mixer;

		adios(1, "system", "$response->{code}: $response->{diagnostic}");
	}
}


# allow our dynamic function calls
no strict 'refs';

# subroutine to invoke
my $func = "big_" . $cmd;

my $status = &$func(@ARGV);

adios($status, "");

# bye
exit 0;


#
# clean up and exit
#
sub adios {
	my $status = shift;
	my $level  = shift;
	my $args   = shift;

	my $result;

	if ($log) {
		if ($level) {
			$log->entry($level, $args);
			print STDERR $args, "\n";
		}

		if ($mixer && ($result = $mixer->fin()) ) {
			$log->entry("error", "mixer::fin", $result);
		}

		$log->entry("info", "end", $status);
		$log->fin();
	}
	elsif ($level) {
		print STDERR $args, "\n";
	}

	exit $status;
}


#
# bulk store
#
sub big_bulk {
	my $subtree = shift;
	my $args    = shift;

	# bulk mode not currently implemented
	$log->entry("info", "bulk mode is not currently supported");

	print STDERR "bulk mode is not currently supported\n";

	return 0;
}



#
# get a block
#
sub big_fetch {
	my($name, $element) = @_;

	# clear global error indicator
	$! = 0;

	my $response = $mixer->fetch($name, $element);

	if ($! == 0) {
		print "$response\n";
		return 0;
	}
	elsif ($! < 0) {
		$log->entry("user", "mixer::fetch",
		            "$response->{code}: $response->{diagnostic}");
		print STDERR "$response->{diagnostic}\n";
		return 1;
	} else {
		$log->entry("error", "mixer::fetch",
		            "$response->{code}: $response->{diagnostic}");

		adios(1, "system", "$response->{code}: $response->{diagnostic}");
	}
}


#
# search for blocks
#
sub big_search {
	my($union, $args) = @_;

	my(%attrs) = ("notification" => "");
	if (defined($args)) {
		%attrs = (%attrs, %{ $args });
	}

	if ($attrs{notification} && ($attrs{notification} ne "false")) {
		print STDERR "notifications not supported\n";
		return 1;
	}

	my $response = $mixer->search($union, \%attrs);

	if ($! == 0) {
		foreach my $name (@{ $response->{answers} }) {
			print "$name\n";
		}
		return 0;
	} elsif ($! < 0) {
		$log->entry("user", "mixer::search",
		            "$response->{code}: $response->{diagnostic}");
		print STDERR "$response->{diagnostic}\n";
		return 1;
	} else {
		$log->entry("error", "mixer::search",
		            "$response->{code}: $response->{diagnostic}");

		adios(1, "system", "$response->{code}: $response->{diagnostic}");
	}
}

#
# put one or more blocks into datastore
#
sub big_store {
	my $subtree = shift;
	my $name    = shift;
	my $action  = shift || "write";
	my $serial  = shift || 1;

	my $reqno  = do_lock($subtree);

	my $status = "";
	if ($reqno) {
		my $input = "";

		if ($action eq "delete") {
			$input = "<block name=" . xml_escape_av($name) . " />";
		}
		else {
			my @input = <STDIN>;
			chomp(@input);
			$input = join("\n", @input);
		}

		$status = do_store($name, $input, $action, $serial);

		do_release($reqno);
	}

	return $status;
}


#
# perform mixer lock
#
sub do_lock {
	my $subtree = shift;

	my $response = $mixer->lock($subtree);

	if ($! == 0) {
		# NO-OP
		;
	} elsif ($! < 0) {
		$log->entry("user", "mixer::lock",
		            "$response->{code}: $response->{diagnostic}");

		adios(1, "user", $response->{diagnostic});
	} else {
		$log->entry("error", "mixer::lock",
		            "$response->{code}: $response->{diagnostic}");

		adios(1, "system", $response->{diagnostic});
	}

	return $response->{reqno};
}


#
# perform mixer store
#
sub do_store {
	my $name   = shift;
	my $body   = shift;
	my $action = shift || "write";
	my $serial = shift || 1;

	my $response = $mixer->store( { name => $name, body => $body }, $action,
	                               $serial);

	if ($! == 0) {
		return 0;
	} elsif ($! < 0) {
		$log->entry("user", "mixer::store",
		            "$response->{code}: $response->{diagnostic}");

		print STDERR "$response->{code}: $response->{diagnostic}\n";
		return 1;
	} else {
		$log->entry("error", "mixer::store",
		            "$response->{code}: $response->{diagnostic}");

		adios(1, "system", "$response->{code}: $response->{diagnostic}");
	}
}


#
# perform mixer lock release
#
sub do_release {
	my $reqno = shift;

	my $response = $mixer->release($reqno);

	if ($! == 0) {
		# NO-OP
		;
	} elsif ($! < 0) {
		$log->entry("user", "mixer::release",
		            "$response->{code}: $response->{diagnostic}");

		print STDERR "$response->{diagnostic}\n";
		return 1;
	} else {
		$log->entry("error", "mixer::release",
		            "$response->{code}: $response->{diagnostic}");

		adios(1, "system", "$response->{code}: $response->{diagnostic}");
	}
}


# noecho for password input
use Term::ReadKey;

# OTP generator
use Space::OTP qw(compute_otp);

#
# passphrase callback
#
sub do_passback {
	my $server    = shift;
	my $name      = shift;
	my $challenge = shift;

	print STDERR "enter secret password for $name\@$server\n$challenge: ";

	# disable echo for password input
	ReadMode 'noecho';
	my $passphrase = ReadLine 0;
	# eat the newline
	chomp $passphrase;
	# restore mode
	ReadMode 'restore';

	print STDERR "\n";

	return compute_otp($challenge, $passphrase, "hex");
}



#
# generate usage info
#
sub usage {
	my $usage = "usage: $0 ?-option value?... command ?arguments...?\n";
	$usage .= "\n";
	$usage .= "options are:\n";

	foreach my $k (sort keys %useropts) {
		$usage .= "    --$k\n";
	}

	$usage .= "\n";
	$usage .= "commands are:\n";
	$usage .= "    fetch name ?element?\n";
	$usage .= "    search union ?arguments...?\n";
	$usage .= "    store subtree name ?action? ?serial?\n";

	$usage;
}


#
# version info
#
sub version {
	return  "Blocks Information Groper " . $VERSION . "r" . $RELEASE;
}


__END__


=head1 NAME

The Blocks Information Groper

=head1 SYNOPSIS

   big.pl ?-option value?... command ?arguments...?

=head1 DESCRIPTION

This program is a front end to the Blocks Mixer.

Arguments appearing before the "command" are key/value pairs as follows.

    -debug boolean:
          indicates that debugging should be enabled.
    -local boolean:
          indicates that the mixer is running on the same system as the server,
          that the user running the mixer has read/write permissions for the
          server's datastore, and the server should be bypassed.
    -bulk boolean:
          indicates that bulk-loading should be used in local mode.
    -mechanism "anonymous":
          indicates that the SASL ANONYMOUS should be used for "authentication".
    -trace string:
          contins the trace information used during the ANONYMOUS SASL exchange.
    -mechanism "otp":
            indicates that the SASL OTP profile should be used for authentication.
    -username string:
          contains the user name used during the OTP SASL exchange.
    -passphrase string:
          contains the passphrase used for OTP SASL exchange.
    -passback upcallV:
         (if no passphrase is supplied) defines an upcall used to return the six
         english words used during the OTP SASL exchange.
    -privacy string:
         the minimum required privacy level (i.e., "none", "weak", or "strong").
    -server serverD:
         the domain name or IP address of the host running the BXXP server
         (default: localhost).
    -logfile file:
         the file name for the log file (default: logs/big.log).

If mixer::init requires a response to an authenticaton challenge, the
"-passback" mechanism is used to query the standard input.


The available commands (and their arguments) are:

=over 3

=item fetch name ?element?

Invokes mixer::fetch and writes the block's body on the standard output.

=item search union ?arguments?

Invokes mixer::search and writes each matching name (on a line by itself) on the
standard output.

=item store subtree name ?action? ?serial?

Locks the subtree, reads a block's body from the standard input, invokes
mixer::store, and unlocks the subtree.

=item bulk subtree ?files...?

Locks the subtree, then reads a block's body from each file invoking
mixer::store, and unlocks the subtree.  If no files are provided on the
command line, then each line of the standard input is read and treated as
the name of the file.

=back

