#!/usr/local/bin/perl
#
#  file: bxxpng.pl
#  auth: brad burdick (bburdick@invisible.net)
#  desc: BXXP ping
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package BXXPng;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use vars qw($VERSION);


# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/examples/client/bxxpng.pl#11 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use
use Fcntl;

# command line option processing
use Getopt::Long;

# generic XML parsing support
use XML::Parser::Expat;

# network socket I/O
use IO::Socket;

# MIME parsing/building support
use MIME::Parser;

# hi-resolution timer support
use Time::HiRes;


# where we find local libraries, if any
use lib "../../lib";


# SpaceKit client support
use Space::Mixer;

# spacekit logging support
use Space::Log;

# spacekit one-time-password support
use Space::OTP qw(MD5_compute);

# spacekit general utility support
use Space::Util qw(boolean timer_log trim);

# spacekit XML support
use Space::XML qw(xml_cdata xml_normalize);


# some useful constants

# other global vars
use vars qw($debug);


# app name
my $prog;
($prog = $0) =~ s|^.*/||;

# debug mode?
$debug = 0;

# log package token
my $log = undef;

# client support
my $mixer = undef;


# default client options
my (%options) = (
	-debug     => "false",
	-generic   => "true",
	-logfile   => "logs/bxxpng.log",
	-mechanism => "none",
	-port      => 10288,
	-privacy   => "none",
	-server    => "sqa.invisible.net",
);


# supported commands and their profile identities
my (%cmds) = (
	echo => "http://xml.resource.org/profiles/NULL/ECHO",
	sink => "http://xml.resource.org/profiles/NULL/SINK",
);


# unique id for ng_null key
my $uid = 0;


#
# main
#

#
# process cmd line options, if any
#
my (%useropts) = (
	debug     => \$options{-debug},
	generic   => \$options{-generic},
	logfile   => \$options{-logfile},
	mechanism => \$options{-mechanism},
	port      => \$options{-port},
	privacy   => \$options{-privacy},
	server    => \$options{-server},
);

my $version = 0;

my $result = GetOptions(\%useropts, "debug=s", "generic=s", "logfile=s",
                          "mechanism=s", "port=i", "privacy=s", "server=s",
                          "version" => \$version);

# display version info and exit
if ($version) {
	print version(), "\n";
	exit 0;
}


if (!$result || ($#ARGV < 0)) {
	die usage();
}


# valid command?
my $cmd;
foreach my $c (keys %cmds) {
	if ($ARGV[0] eq $c) {
		$cmd = pop(@ARGV);
		last;
	}
}

unless ($cmd) {
	adios(1, "user", "unknown command: $ARGV[0]");
}


# global debug mode
$debug = boolean($options{-debug});

# log token
my $ident = (split(/\./, $prog))[0];

$log = Space::Log->new($options{-logfile}, $ident);
unless ($log) {
	adios(1, "user", "log::new failed");
}

$log->entry("info", "begin $cmd");

my $server = $options{-server};

# remove local config options
delete $options{-logfile};
delete $options{-server};

$mixer = new Space::Mixer(); 
if ( ! $mixer || ($result = $mixer->init($log, $server, \%options)) ) {
	$log->entry("error", "mixer::init",
	            "$result->{code}: $result->{diagnostic}");

	$mixer = undef;
	adios(1, "system", "$result->{code}: $result->{diagnostic}");
}


# need to disable strict refs for func calls below
no strict 'refs';

# subroutine to invoke
my $func = "bxxpng_" . $cmd;

my $status = &$func(@ARGV);

adios($status, "");


# c'ya
exit 0;



#
#
#

#
# clean up and exit
#
sub adios {
	my $status = shift;
	my $level  = shift;
	my $args   = shift;

	my $result;

	if ($log) {
		if ($level) {
			$log->entry($level, $args);
			print STDERR $args, "\n";
		}

		if ($mixer && ($result = $mixer->fin()) ) {
			$log->entry("error", "mixer::fin", $result);
		}

		$log->entry("info", "end", $status);
		$log->fin();
	}
	elsif ($level) {
		print STDERR $args, "\n";
	}

	if ($debug > 1) {
		print STDERR "exit $status\n";
		return 0;
	}

	exit $status;
}


#
# BXXP ping
#

sub bxxpng_echo {
	my $args = shift;

	return ng_null("echo", $args);
}


sub bxxpng_sink {
	my $args = shift;

	return ng_null("sink", $args);
}


sub ng_null {
	my $mode = shift;
	my $opts = shift;

	# set some default options -- merge the rest
	my %params = ( -count => 5, -size => 1024 );
	%params = ( %params, %{ $opts } );

	my $echo = 0;
	my $profile = "";
	my $channel = 0;

	if ($mode eq "echo") {
		$echo = 1;
	}
	elsif ($mode eq "sink") {
		$echo = 0;
	}
	$profile = $cmds{$mode};

	my $hit = 0;
	foreach my $uri (keys %{ $mixer->profiles() }) {
		if ($profile eq $uri) {
			$hit = 1;
			last;
		}
	}

	unless ($hit) {
		$log->entry("user", "$profile not advertised");
	}

	my $result = $mixer->create($profile);
	if ($! == 0) {
		# ok
		$channel = $result;
		$log->entry("debug", $mode, "started on channel $channel");
	}
	elsif ($! < 0) {
		# internal error
		croak("$result->{code}: $result->{diagnostic}");
	}
	else {
		# system error
		croak("$result->{code}: $result->{diagnostic}");
	}

	my $forever;
	if ($params{-count} < 0) {
		$forever = 1;
	}
	elsif ($params{-count} == 0) {
		return 0;
	}
	else {
		$forever = 0;
	}

	$params{-size} -= 13;
	if ($params{-size} < 0) {
		$params{-size} = 0;
	}

	print STDERR "initializing buffer... ";

	use Sys::Hostname;
	my $host = eval { hostname(); };

	my $id = $uid++ . "\@" . $host;
	my $key = join(".", $$, time(), $id);

	my $data = "";
	while (length($data) < $params{-size}) {
		my $x = unpack("H*", ($key = MD5_compute($key, 1)));
		$data .= join(" ", $x x 5) . "\n";
	}

	print STDERR "done.\n";

	my $status = 0;
	my $min = 0;
	my $avg = 0;
	my $max = -1;
	my $cnt = $params{-count};

	my $nlP = 0;

	# don't buffer STDOUT
	select STDOUT; $| = 1;

	my ($req, $rsp);
	while ($forever || ($params{-count}-- > 0)) {
		if ($params{-size} > 1) {
			# roll data
			$data =~ s/^(.)(.*)$/$2$1/gs;
		}

		$req = build MIME::Entity Type     => "text/xml",
		                          Encoding => "8bit",
		                          Data     => "<data>$data</data>";

		my $click = Time::HiRes::time();

		$rsp = $mixer->exch($channel, $req);
		if ($! == 0) {
			# NO-OP
			;
		}
		elsif ($! < 0) {
			# internal error
			my $result = $mixer->errscan($rsp);
			croak("$result->{code}: $result->{diagnostic}");
		}
		else {
			# system error
			croak("$rsp->{code}: $rsp->{diagnostic}");
		}

		$click = Time::HiRes::time() - $click;

		if ($max <= 0) {
			$min = $max = sprintf("%6.6f", $click);
		}
		else {
			if ($click < $min) {
				$min = sprintf("%6.6f", $click);
			}

			if ($click > $max) {
				$max = sprintf("%6.6f", $click);
			}
		}

		$avg += sprintf("%6.6f", $click);

		if ($echo && ($req->stringify_body() ne $rsp->stringify_body())) {
			if ($nlP) {
				print "\n";
				$nlP = 0;
			}

			print STDERR "$channel: data mismatch\n";
			$status = 1;
		}

		print ".";
		if ($nlP++ > 78) {
			print "\n";
			$nlP = 0;
		}

		# release MIME::Entity
		undef $req;
		undef $rsp;
	}

	if ($nlP) {
		print "\n";
	}

	$avg = $avg/$cnt;
	my $prefix = "$cnt round-trips: min/avg/max =";
	print sprintf("%s %4.3f %4.3f %4.3f (ms)\n", $prefix,
	                $min*1000, $avg*1000, $max*1000);

	if ($params{-size} > 0) {
		my $len = length($prefix);

		$prefix = sprintf("%*.*s", $len, $len, "");
		print $prefix;

		my $speed;
		foreach my $secs (($min, $avg, $max)) {
			if ($secs > 0) {
				$speed = int($params{-size}/$secs) >> 10;
			}
			else {
				$speed = "";
			}

			print sprintf("%8.8s", $speed);
		}

		print " (KB/s)\n";
	}

	return $status;
}


#
# display usage info
#
sub usage {
	my $usage = "usage: $prog ?-option value?... command ?arguments...?\n";
	$usage .= "\n";
	$usage .= "commands are:\n";

	foreach my $c (keys %cmds) {
		$usage .= "\t$c ?arguments...?\n";
	}

	return $usage;
}


#
# version info
#
sub version {
	return  "BXXP Ping " . $VERSION . "r" . $RELEASE;
}


__END__

