#!/usr/local/bin/perl
#
#  file: bxxd.pl
#  auth: brad burdick (bburdick@invisible.net)
#  desc: BXXD - Blocks server implementation 
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package BXXD;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use vars qw($VERSION);


# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/examples/server/bin/bxxd.pl#11 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# command line option processing
use Getopt::Long;

# for peer address info
use Socket;

# object dump support
use Data::Dumper;

# configuration tool
use App::Config;

# generic network server support (for daemon mode)
use NetServer::Generic;

# XML parsing support
use XML::Simple;


# spacekit blocks server support
use Space::Server;

# spacekit logging support
use Space::Log;

# spacekit general utility support
use Space::Util qw(timer_init timer_start timer_fin);


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;


# app name
my $prog;
($prog = $0) =~ s|^.*/||;

# log package object
my $log = undef;


#
# main
#


# BXXD config file
my $cfg_file = "";

# am i running as a standalone server?  default is to run from INETD.
my $daemon = 0;

# display version info?
my $version = 0;

# default daemon port to use -- not currently used as there is only a single
# bxxd configuration available with this simple server implementation.  it
# would normally specify which bxxd configuration to use from the config file.
my $port = 10288;


# get cmd line options, if any
GetOptions("config=s" => \$cfg_file, "daemon" => \$daemon, "port" => \$port,
            "version" => \$version);

# display version info and exit
if ($version) {
	print version(), "\n";
	exit 0;
}


# default config file if not provided
$cfg_file = "etc/bxxd-config.xml"
	unless ($cfg_file);


# load BXXD config
my $cfg = load_config($cfg_file);

# config specified debug level
my $debuglevel = $cfg->debugLevel();


# run relative to homeDirectory if available, /usr/local/bxxd if not
my $homeDir = $cfg->get("homeDirectory");
$homeDir = "/usr/local/bxxd" unless ($homeDir);

chdir $homeDir;


# initialise log environment
$log = Space::Log->new($cfg->logFile(), "bxxd", "");


# running in daemon mode?  default is INETD
if ($daemon) {
	my $netserver = NetServer::Generic->new();

	# my listener port
	$netserver->port($cfg->port());
	# daemon mode to use: forking, select
	$netserver->mode("forking");
	$netserver->callback(\&bxxd);

	$netserver->run();
}
else {
	bxxd();
}


# c'ya
adios(0);


#
# space server
#
sub bxxd {
	# NetServer object
	my $netserver = shift;

	# get client info
	my %client = ();
	if ($debug) {
		%client = (
			address => "debug",
			domain  => "",
			port    => "",
		);
	}
	else {
		my $remote         = getpeername(STDIN);
		my ($port, $iaddr) = unpack_sockaddr_in($remote);
		my $remote_host    = gethostbyaddr($iaddr, AF_INET);
		my $remote_addr    = inet_ntoa($iaddr);

		%client = (
			address => $remote_addr,
			domain  => $remote_host,
			port    => $port,
		);
	}


	# log client activity
	$log->entry("info", "begin", $client{address});


	# load available profiles
	load_profiles($cfg);


	my $server = new Space::Server();
	if ($server) {
		# initialise server state
		$server->init($log, \%client, \*STDIN, \*STDOUT, $cfg->profiles());

		# let's get busy
		my $result = $server->run();

		if ($result) {
			$server->fin(1);
			adios(1, "user", $result->{diagnostic});
		}

		# ok finish status
		$server->fin(0);
	}
	else {
		adios(1, "fatal", "error creating new server instance!");
	}

	0;
}


#
# clean up and exit
#
sub adios {
	my $status = shift;
	my $level  = shift;
	my $args   = shift || "";

	if ($log) {
		if ($level) {
			$log->entry($level, $args);
			print STDERR $args, "\n"
				if ($debug);
		}

		$log->entry("info", "end", $status);
		$log->fin();
	}

	if ($debug > 1) {
		print STDERR "exit $status\n";
		return 0;
	}

	exit $status;
}


#
# load BXXD config (defaults and file)
#
sub load_config {
	my $file = shift;

	# global bxxd configuration data
	my $cfg = App::Config->new( { CASE => 1, FILEPARSE => \&_read_config_file } );

	# start with default config settings
	set_default_config($cfg);

	# load bxxd config -- uses _read_config_file()
	my $ret = $cfg->cfg_file($file);

	print STDERR "load: ", Dumper($cfg), "\n"
		if ($debug > 1);

	$cfg;
}


#
# can't have strict refs with dynamic function names
no strict 'refs';

#
# load available profile modules
#
sub load_profiles {
	my $cfg = shift;

	# where we find our modules
	my $moduleDir = $cfg->moduleDirectory();
	unless ($moduleDir =~ m#^/#) {
		$moduleDir = join("/", $cfg->homeDirectory(), $moduleDir);
	}
	unshift(@INC, $moduleDir);

	# get local copy of profiles to update
	my %profiles = %{$cfg->profiles()};

PROFILE:
	foreach my $profile (keys %profiles) {
		if (exists($profiles{$profile}->{module})) {
			eval { require "Profile/" . $profiles{$profile}->{module} . ".pm"; };
			if ($@) {
				# error loading profile?  remove the profile from list
				print STDERR "$profile: error loading module: $@!\n";
				delete $profiles{$profile};
				next PROFILE;
			}

			# get profile info

			my $info_func = "Profile::" . $profiles{$profile}->{module} . "::info";
			# convert any filename chars that will cause problems
			$info_func =~ s/-/_/g;

			# get info array ref
			my $info = &$info_func($log);

			# make sure this profile identity matches what we want
			my $match = 0;
			foreach my $id (@{$info->[1]}) {
				if ($id eq $profile) {
					$match = 1;
					last;
				}
			}

			unless ($match) {
				print STDERR "$profile: does not match any module identity!\n";
				delete $profiles{$profile};
				next PROFILE;
			}

			# does this profile provide transport security (encryption)?
			$profiles{$profile}->{privProvider} = $info->[0];

			# profile public methods to use
			my $methods = $info->[2];
			foreach my $k (keys %{$methods}) {
				$profiles{$profile}->{$k} = $methods->{$k};
			}
		}
		else {
			# no module to load?  remove the profile from list
			print STDERR "$profile: missing module to load!\n";
			delete $profiles{$profile};
			next PROFILE;
		}
	}

	# update list of available profiles
	$cfg->set('profiles', { %profiles });

	0;
}


#
# set default configuration values
#
sub set_default_config {
	my $cfg = shift;

	# default configuration settings for global parameters
	my %default_config = (
		authDirectory   => "auth",
		debugLevel      => "none",
		homeDirectory   => "/usr/local/bxxd",
		logFile         => "logs/bxxd.log",
		moduleDirectory => "lib",
		port            => 10288,
	);

	while (my($k, $v) = each %default_config) {
		$cfg->define($k);
		$cfg->set($k, $v);
	}

	0;
}


#
# read BXXD config file
#
sub _read_config_file {
	my $cfg = shift;
	my $file = shift;

	unless (-r $file) {
		# unable to read config file, then error
		return 0;
	}

	my $config = eval { XMLin($file); };

	#
	# set config options
	#
	$cfg->set("port", $config->{bxxd}->{port})
		if ($config->{bxxd}->{port});

	my $param_href = $config->{bxxd}->{parameter};
	foreach my $k (keys %{ $param_href }) {
		print STDERR "param: $k => $param_href->{$k}->{value}\n"
			if ($debug > 2);
		$cfg->define($k);
		$cfg->set($k, $param_href->{$k}->{value})
	}

	my %profs;
	my $prof_aref = $config->{bxxd}->{profile};
	foreach my $h (@{ $prof_aref }) {
		my %prof;
		foreach my $k (keys %{ $h }) {
			if (ref($h->{$k}) eq "HASH") {
				print STDERR "profs: params: $h->{$k}->{name} => $h->{$k}->{value}\n"
					if ($debug > 2);
				$prof{$h->{$k}->{name}} = $h->{$k}->{value};
			}
			else {
				print STDERR "profs: $k => $h->{$k}\n"
					if ($debug > 2);
				$prof{$k} = $h->{$k};
			}
		}

		$profs{$prof{uri}} = { %prof };
	}

	if (keys(%profs)) {
		$cfg->define("profiles");
		$cfg->set("profiles", { %profs });
	}

	# App::Config requires 1 response for success
	1;
}


#
# display usage info
#
sub usage {
	my $usage = "usage: $prog [--config path/to/config]\n";

	return $usage;
}


#
# version info
#
sub version {
	return __PACKAGE__ . " " . $VERSION . "r" . $RELEASE;
}


__END__

