#!/usr/local/bin/perl
#
#  file: sasl-adm.pl
#  auth: brad burdick (bburdick@invisible.net)
#  desc: SASL admin support
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package SASLadmin;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use vars qw($VERSION);


# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/examples/server/etc/sasl-adm.pl#10 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# command line option processing
use Getopt::Long;

# password input mode support
use Term::ReadKey;


# where we find local libraries, if any
use lib "./lib";

# spacekit logging support
use Space::Log;

# spacekit One-Time-Password support
use Space::OTP qw(SHA1_compute hex_to_word);

# spacekit SASL support
use Space::SASL;


# other global vars
use vars qw($debug);


# debug level?
$debug = 0;


# app name
my $prog;
($prog = $0) =~ s|^.*/||;


#
# main
#

# default umask for file/directory creation
umask 0002;


# action to perform (create/update/delete/list), default to create
my $action = "create";

# BXXD (default) authdb directory, ./auth
my $authdir = "auth";

# log file, with default
my $logfile = "logs/sasl-adm.log";

# SASL mechanism to use (anonymous/otp), default to OTP
my $mechanism = "otp";

# display version?
my $version = 0;

# get cmd line options, if any
GetOptions("action=s"    => \$action,
           "authdir=s"   => \$authdir,
           "logfile=s"   => \$logfile,
           "mechanism=s" => \$mechanism,
           "version"     => \$version
          );


if ($version) {
	print version(), "\n";
	exit 0;
}


# valid options?
if ($action !~ /^(create|delete|list|update)$/i) {
	die "$prog: unsupported action: $action\n";
}

if ($mechanism && ($mechanism !~ /^(anonymous|otp)$/i)) {
	die "$prog: unsupported mechanism: $mechanism\n";
}


# username to work on
my $username = shift || "anonymous";

# does authdir exist?  prompt to create it if not
unless (-d $authdir) {
	print STDERR "$authdir does not exist.  create it? ";

	my $have_authdir = 0;
	if (<STDIN> =~ /^(y|yes)$/i) {
		# let umask control user/group perms
		$have_authdir = mkdir $authdir, 0770;
	}

	unless ($have_authdir) {
		die "cannot proceed without auth directory...aborting!\n";
	}
}


# remaining arguments should be subtrees you want write access to
my $subtrees = [ @ARGV ];


my $status = 1;

my $log = Space::Log->new($logfile, "sasl-adm");
$log->entry("info", "begin", "$action $username");

my $sasl = Space::SASL->new();

unless ($sasl) {
	die "$prog: error getting SASL access: $!\n";
}
	
$sasl->init($log, "", "", $authdir);


# clear global error indicator
$! = 0;

my $auth;

if ($mechanism eq "anonymous") {
	if ($action eq "create") {
		$auth = anon_create($log, $sasl, $subtrees);
	}
	elsif ($action eq "delete") {
		$auth = undef;
	}
	elsif ($action eq "update") {
		# can only change subtrees...for now
		$auth = anon_update($log, $sasl, $subtrees);
	}
}
elsif ($mechanism eq "otp") {
	if ($action eq "create") {
		$auth = otp_create($log, $sasl, $username, $subtrees);
	}
	elsif ($action eq "delete") {
		$auth = undef;
	}
	elsif ($action eq "update") {
		# can only change subtrees...for now
		$auth = otp_update($log, $sasl, $username, $subtrees);
	}
}

# list requested username SASL info
if ($action eq "list") {
	$auth = list_user($sasl, $username ? $username : "anonymous");
}
else {
	if ($! == 0) {
		# update SASL data
		$sasl->lock($username);

		$sasl->store($username, $auth, $action);

		$sasl->release($username);
	}
}

if ($! == 0) {
	# ok
	print "done.", "\n";

	$status = 0;
	$log->entry("info", "end", $status);
}
elsif ($! < 0) {
	# internal error
	print STDERR "$auth->{code}: $auth->{diagnostic}\n";

	$log->entry("user", "$auth->{code}: $auth->{diagnostic}");
	$log->entry("info", "end", $status);
}
else {
	# system error
	$log->entry("system", $!);

	print STDERR $!, "\n";

	$log->entry("info", "end", $status);
}


# release SASL instance
$sasl->fin();

# release logging instance
$log->fin();

# c'ya
exit $status;


#
# create the ANONYMOUS user
#
sub anon_create {
	my $log      = shift;
	my $sasl     = shift;
	my $subtrees = shift;

	# clear global error indicator
	$! = 0;

	my $result = $sasl->fetch("anonymous");
	if ($! == 0) {
		# user exists...error
		$! = -7;

		return { code => 554, diagnostic => "anonymous account already exists" };
	}
	elsif ($! < 0) {
		# fetch failed...user doesn't exist yet
		if ($result->{code} != 550) {
			$! = -7;
			return $result;
		}
	}
	else {
		# system error
		return $result;
	}

	# build acls
	my %acls;

	# default acl
	$acls{""} = { privs => "fetch", subtree => "" };

	foreach my $subtree (@{ $subtrees }) {
		$acls{$subtree} = {
			privs   => "create delete fetch lock notify write update",
			subtree => $subtree,
		};
	}

	# authentication data
	my %auth = (
		name => "anonymous",
		authentication => {
			mechanism => "anonymous",
			privacy   => "none",
		},
		authorization => {
			acl => { %acls },
		},
	);

	print "Create ID anonymous\n";

	$log->entry("info", "create", "anonymous");

	# clear global error indicator
	$! = 0;

	return { %auth };
}


#
# update the ANONYMOUS user
#
sub anon_update {
	my $log      = shift;
	my $sasl     = shift;
	my $subtrees = shift;

	# clear global error indicator
	$! = 0;

	my $result = $sasl->fetch("anonymous");
	if ($! == 0) {
		# ok
		;
	}
	elsif ($! < 0) {
		# fetch failed...user doesn't exist yet
		$! = -7;

		return { code => 554, diagnostic => "anonymous account does not exist" };
	}
	else {
		# system error
		return $result;
	}

	# build acls
	foreach my $subtree (@{ $subtrees }) {
		$result->{authorization}->{acl}->{$subtree} = {
			privs   => "create delete fetch lock notify write update",
			subtree => $subtree,
		};
	}

	print "Update ID anonymous\n";

	$log->entry("info", "update", "anonymous");

	# clear global error indicator
	$! = 0;

	return $result;
}


#
# create a new OTP user
#
sub otp_create {
	my $log      = shift;
	my $sasl     = shift;
	my $name     = shift;
	my $subtrees = shift;

	# clear global error indicator
	$! = 0;

	my $result = $sasl->fetch($name);
	if ($! == 0) {
		# user exists...error
		$! = -7;

		return { code => 554, diagnostic => "account $name already exists" };
	}
	elsif ($! < 0) {
		# fetch failed...user doesn't exist yet
		if ($result->{code} != 550) {
			$! = -7;
			return $result;
		}
	}
	else {
		# system error
		return $result;
	}

	# get passphrase from user
	my $pass = otp_passphrase($name);
	if ($! < 0) {
		return $pass;
	}

	# generate OTP info
	my $seq  = 9999;
	my $seed = Space::OTP::makeseed();

	my $key = SHA1_compute("$seed$pass", $seq);

	# build acls
	my %acls;

	# default acl
	$acls{""} = { privs => "fetch", subtree => "" };

	foreach my $subtree (@{ $subtrees }) {
		$acls{$subtree} = {
			privs   => "create delete fetch lock notify write update",
			subtree => $subtree,
		};
	}

	# authentication data
	my %auth = (
		name => $name,
		authentication => {
			algorithm => "sha1",
			key       => $key,
			mechanism => "otp",
			privacy   => "none",
			seed      => $seed,
			sequence  => $seq,
		},
		authorization => {
			acl => { %acls },
		},
	);

	print "Create ID $name otp is $seq $seed\n" . hex_to_word($key) . "\n";

	$log->entry("info", "create", "$name $seq $seed");

	# clear global error indicator
	$! = 0;

	return { %auth };
}


#
# update an existing OTP user
#
sub otp_update {
	my $log      = shift;
	my $sasl     = shift;
	my $name     = shift;
	my $subtrees = shift;

	# clear global error indicator
	$! = 0;

	my $result = $sasl->fetch($name);
	if ($! == 0) {
		# ok
		;
	}
	elsif ($! < 0) {
		# fetch failed...user doesn't exist
		$! = -7;

		return { code => 554, diagnostic => "account $name does not exist" };
	}
	else {
		# system error
		return $result;
	}

	# get passphrase from user
	my $pass = otp_passphrase($name);
	if ($! < 0) {
		return $pass;
	}

	# generate OTP info
	my $seq  = $result->{authentication}->{sequence};
	my $seed = $result->{authentication}->{seed};

	my $key = SHA1_compute("$seed$pass", $seq);

	# invalid passphrase?
	unless ($key eq $result->{authentication}->{key}) {
		$! = -7;
		return { code => 554, diagnostic => "invalid passphrase" };
	}

	# update acls
	foreach my $subtree (@{ $subtrees }) {
		$result->{authorization}->{acl}->{$subtree} = {
			privs   => "create delete fetch lock notify write update",
			subtree => $subtree,
		};
	}

	print "Update ID $name otp is $seq $seed\n" . hex_to_word($key) . "\n";

	$log->entry("info", "update", "$name $seq $seed");

	# clear global error indicator
	$! = 0;

	return $result;
}


#
# get OTP passphrase from user
#
sub otp_passphrase {
	my $name = shift;

	my ($pass1, $pass2);

	for (my $i = 0 ; 1 ; $i++) {
		print STDERR "Enter passphrase for $name: ";

		# disable echo for password input
		ReadMode 'noecho';

		$pass1 = ReadLine 0;
		# eat the newline
		chomp $pass1;

		print STDERR "\nRe-enter passphrase for $name: ";

		$pass2 = ReadLine 0;
		# eat the newline
		chomp $pass2;

		# restore mode
		ReadMode 'restore';

		if ($pass1 eq $pass2) {
			my $len = length($pass1);
			if ($len < 10 || $len > 63) {
				$! = -7;
				return { code => 501, diagnostic => "passphrase must have 10..63 characters" };
			}

			# good passphrase, we're done
			last;
		}

		print STDERR "Sorry, no match\n";
		if ($i >= 2) {
			$! = -7;
			return { code => 500, diagnostic => "passphrase input failed" };
		}
	}

	print "\n";

	# clear global error indicator
	$! = 0;

	return $pass1;
}


#
# list requested user'd SASL info
#
sub list_user {
	my $sasl = shift;
	my $name = shift;

	# clear global error indicator
	$! = 0;

	my $result = $sasl->fetch($name);
	if ($! == 0) {
		print "\n";
		_dump_auth($result, 0);
		print "\n";
	}
	elsif ($! < 0) {
		# fetch failed...user doesn't exist
		$! = -7;

		return { code => 554, diagnostic => "account $name does not exist" };
	}
	else {
		# system error
		return $result;
	}

	# clear global error indicator
	$! = 0;

	return {};
}


#
# display usage info
#
sub usage {
	my $usage = "usage: $prog\n";
	$usage .= "  [--action create|delete|list|update]\n";
	$usage .= "  [--authdir directory]\n";
	$usage .= "  [--logfile file]\n";
	$usage .= "  [--mechanism anonymous|otp]\n";
	$usage .= "  [--version]\n";
	$usage .= "  [username [subtree ...]]\n";

	return $usage;
}


#
# return program version info
#
sub version {
	return "BXXD SASL Administration tool, ${VERSION}r${RELEASE}";
}


#
# dump SASL data structure
#
sub _dump_auth {
	my $auth = shift;
	my $indent = shift;

	my $s = " " x $indent;

	foreach my $k (sort keys %{$auth}) {
		my $l = $k ? $k : '""';

		if (ref($auth->{$k}) eq "HASH") {
			print "$s$l => {", "\n";
			_dump_auth($auth->{$k}, $indent+2);
			print "$s}", "\n";
		}
		else {
			my $m = $auth->{$k};
			$m = qq|"$m"|
				if ($m =~ /\s/);
			print "$s$l => $m", "\n";
		}
	}

	0;
}


__END__

