#
#  file: Profile/sasl-anon.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: SASL/ANON (anonymous) profile support
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Profile::sasl_anon;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw();

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/examples/server/lib/Profile/sasl-anon.pm#7 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use


# spacekit SASL support
use Space::SASL;

# spacekit logging support
use Space::Log;

# spacekit general XML support
use Space::XML qw(xml_cdata);


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;



#
# state variables
#
# log:  object for logging package
#


#
# anon profile boot protocol (run-time bootable?)
#
sub boot {
	my $log = shift;

	# clear global error indicator
	$! = 0;

	# is SASL anonymous supported and loadable?
	my $sasl = Space::SASL->new();

	$sasl->init($log, "", "anonymous")
		if ($sasl);

	if ($! || ! $sasl) {
		$log->entry("error", "sasl::init", $!);
		return { code => 535, diagnostic => "authentication failed" };
	}

	$sasl->fetch("anonymous");
	# non-internal error of some kind?
	if ($! > 0) {
		$log->entry("error", "sasl::fetch", $!);

		# change to internal error code
		$! = -7;

		return { code => 451, dignostic => $! };
	}

	if ($sasl->fin()) {
		$log->entry("error", "sasl::fin", $!);

		# internal error code
		$! = -7;

		return { code => 535, diagnostic => "authentication failed" };
	}

	0;
}


#
# anon profile info
#
sub info {
	my $log = shift;

	return [ 0,  # profile provides transport security? (0 or 1)
	         [   # supported profile identities (an XML DTD)
	           "http://xml.resource.org/profiles/sasl/ANONYMOUS"
	         ],
	         {   # public methods of this profile
	           bootV => \&Profile::sasl_anon::boot,
	           initV => \&Profile::sasl_anon::init,
	           exchV => \&Profile::sasl_anon::exch,
	           finV  => \&Profile::sasl_anon::fin
	         }
	       ];
}


#
# anon profile initialisation
#
sub init {
	my $log     = shift;
	my $serverD = shift;
	my $clientA = shift;
	my $upcallV = shift;
	my $uri     = shift;

	# clear global error
	$! = 0;

	my $sasl = Space::SASL->new();
	$sasl->init($log, $clientA, "anonymous")
		if ($sasl);

	unless ($sasl) {
		$log->entry("error", "sasl::init", $!);
		return {};
	}

	return { log => $log, sasl => $sasl };
}


#
# anon profile finalisation
#
sub fin {
	my $token  = shift;
	my $status = shift;

	my $result = $token->{sasl}->fin();
	if ($!) {
		$token->{log}->entry("error", "sasl::fin", $result);
	}

	# clean up profile token
	foreach my $k (keys %{$token}) {
		$token->{$k} = undef;
	}

	$token = undef;

	0;
}


#
# anon profile data/message exchange
#
sub exch {
	my $token = shift;
	my $data  = shift;

	# clear global error state
	$! = 0;

	my $result = $token->{sasl}->parse($data);
	if ($! == 0) {
		return exch1($token, $result);
	}
	elsif ($! < 0) {  ## internal error indicator
		my $err = "<error code='$result->{code}'>\n";
		$err .= xml_cdata($result->{diagnostic});
		$err .= "</error>";

		return $err;
	}
	else {            ## system/perl error
		$token->{log}->entry("error", "sasl::parse", $result);

		return $result;
	}

	# NOT-REACHED
	0;
}


#
# anon profile data/message exchange part deux
#
sub exch1 {
	my $token = shift;
	my $data  = shift;

	# clear global error indicator
	$! = 0;

	my $trace = $data->{authenticator};
	my $other = $data->{authorization};

	if ($other && $other ne "anonymous") {
		$! = -7;
		my $err = "<error code='551'>\n";
		$err .= xml_cdata("proxying not supported");
		$err .= "</error>";

		return $err;
	}

	my $result = $token->{sasl}->fetch("anonymous");
	if ($! == 0) {
		# NO-OP
		;
	}
	elsif ($! < 0) {  ## internal error indicator
		if ($result->{code} == 550) {
			$result->{diagnostic} = "anonymous login not allowed";
		}

		my $err = "<error code='$result->{code}'>\n";
		$err .= xml_cdata($result->{diagnostic});
		$err .= "</error>";

		return $err;
	}
	else {
		$token->{log}->entry("error", "sasl::fetch", $result);

		return $result;
	}

	# valid authenticator info supplied?
	if ( ($trace !~ /\@/) && (length($trace) > 255) ) {
		my $err = "<error code='501'>\n";
		$err .= xml_cdata("poorly formed <authenticator> element");
		$err .= "</error>";

		return $err;
	}

	# complete SASL login
	$token->{sasl}->login("anonymous", { trace => $trace });

	return "<complete />";
}


#
# library version info
#
sub version {
	return __PACKAGE__ . " " . $VERSION . "r" . $RELEASE;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

