#
#  file: Profile/sasl-otp.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: SASL/OTP (one-time-password) profile support
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Profile::sasl_otp;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw();

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/examples/server/lib/Profile/sasl-otp.pm#7 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# MIME Base64 encode/decode support
use MIME::Base64;


# spacekit SASL support
use Space::SASL;

# perl extended OTP support
use Space::OTP qw(compute_otp makeseed);

# spacekit logging support
use Space::Log;

# spacekit general XML support
use Space::XML qw(xml_cdata);


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;


#
# state variables
#
#  log:     object for logging package
#  sasl:    object for sasl package
#  rspP:    expecting <response> element
#  account: user information
#  loseP:   sent losing <challenge> (on purpose)
#  updateP: need to update SASL database
#
# present only during <response> parsing
#
#  code:       reply code
#  diagnostic: informative text
#


#
# otp profile boot protocol (run-time bootable?)
#
sub boot {
	my $log = shift;

	# clear global error indicator
	$! = 0;

	# make sure OTP support is available/working
	unless ($Space::OTP::VERSION) {
		$! = -7;

		return { code => 451, diagnostic => "Space::OTP module missing or incomplete" };
	}

	0;
}


#
# otp profile info
#
sub info {
	my $log = shift;

	return [ 0,  # profile provides transport security? (0 or 1)
	         [   # supported profile identities (an XML DTD)
	           "http://xml.resource.org/profiles/sasl/OTP"
	         ],
	         {   # public methods of this profile
	           bootV => \&Profile::sasl_otp::boot,
	           initV => \&Profile::sasl_otp::init,
	           exchV => \&Profile::sasl_otp::exch,
	           finV  => \&Profile::sasl_otp::fin
	         }
	       ];
}


#
# otp profile initialisation
#
sub init {
	my $log     = shift;
	my $serverD = shift;
	my $clientA = shift;
	my $upcallV = shift;
	my $uri     = shift;

	# clear global error
	$! = 0;

	my $sasl = Space::SASL->new();
	$sasl->init($log, $clientA, "otp")
		if ($sasl);

	unless ($sasl) {
		$log->entry("error", "sasl::init", $!);
		return {};
	}

	return { log => $log, sasl => $sasl, rspP => 0,
	         account => (), loseP => 0, updateP => 0 };
}


#
# otp profile finalisation
#
sub fin {
	my $token  = shift;
	my $status = shift;

	# clear global error indicator
	$! = 0;

	my $result = $token->{sasl}->fin();
	if ($!) {
		$token->{log}->entry("error", "sasl::fin", $result);
	}

	# clean up profile token
	foreach my $k (keys %{$token}) {
		$token->{$k} = undef;
	}

	$token = undef;

	0;
}


#
# otp profile data/message exchange
#
sub exch {
	my $token = shift;
	my $data  = shift;

	# clear global error indicator
	$! = 0;

	if (! $token->{rspP}) {
		$token->{account} = ();
		$token->{loseP} = 0;
		$token->{updateP} = 0;
	}

	my $result = $token->{sasl}->parse($data);
	if (! $!) {
		if ($token->{rspP}) {
			return exch2($token, $result);
		}
		else {
			return exch1($token, $result);
		}
	}
	elsif ($! < 0) {
		$token->{rspP} = 0;
		my $err = "<error code='$result->{code}'>\n";
		$err .= xml_cdata($result->{diagnostic}) . "\n";
		$err .= "</error>";

		return $err;
	}
	else {
		$token->{log}->entry("error", "sasl::parse", $!);
		$token->{rspP} = 0;

		$! = -7;
		return $result;
	}

	# NOT-REACHED
	0;
}


#
# otp profile data/message exchange part 2
#
sub exch1 {
	my $token = shift;
	my $data  = shift;

	# clear global error indicator
	$! = 0;

	my $name = $data->{authenticator};
	my $other = $data->{authorization};
	if ($other && ($name ne $other)) {
		my $err = "<error code='551'>\n";
		$err .= xml_cdata("proxying not supported") . "\n";
		$err .= "</error>";
		return $err;
	}

	my %otp = ();

	my $result = $token->{sasl}->lock($name);
	if (! $!) {
		# NO-OP
		;
	}
	elsif ($! < 0) {
		my $err = "<error code='$result->{code}'>\n";
		$err .= xml_cdata($result->{diagnostic}) . "\n";
		$err .= "</error>";

		$! = -7;
		return $err;
	}
	else {
		my $err = $!;
		$token->{log}->entry("error", "sasl::lock", $err);

		$! = -7;
		return $err;
	}
	
	$result = $token->{sasl}->fetch($name);
	if (! $!) {
		$token->{account} = $result;
		%otp = %{ $result->{authentication} };
	}
	elsif ($! < 0) {
		if ($result->{code} != 550) {
			my $err = "<error code='$result->{code}'>\n";
			$err .= xml_cdata($result->{diagnostic}) . "\n";
			$err .= "</error>";

			release($token, $name);
			$! = -7;
			return $err;
		}

		$token->{loseP} = 1;
		%otp = (
			algorithm => "md5",
			key       => 0,
			seed      => makeseed(),
			sequence  => "1" . (time() % 1000),
		);

		$token->{account} = { name => $name, authentication => { %otp } };
	}
	else {
		my $err = $!;
		$token->{log}->entry("error", "sasl::fetch", $!);

		$! = -7;
		return $err;
	}

	if ($otp{sequence} <= 1) {
		my $err = "<error code='534'>\n";
		$err .= xml_cdata("sequence exhausted") . "\n";
		$err .= "</error>";

		release($token, $name);

		$! = -7;
		return $err;
	}

	# we have a challenge response
	$token->{rspP} = 1;
	$data = "otp-$otp{algorithm} " . $otp{sequence}-1 . "$otp{seed} ext";

	return "<challenge>" . xml_cdata($data) . "</challenge>";
}


#
# release lock for specified username
#
sub release {
	my $token = shift;
	my $name  = shift;

	# clear global error indicator
	$! = 0;

	my $result = $token->{sasl}->release($name);
	if (! $!) {
		# NO-OP
		;
	}
	elsif ($! < 0) {
		$token->{log}->entry("notify", "sasl::release", $result->{code}, ": ",
		            $result->{diagnostic});
	}
	else {
		$token->{log}->entry("notify", "sasl::release", $!);
	}

	0;
}


#
# otp profile data/message exchange part 3
#
sub exch2 {
	my $token = shift;
	my $data  = shift;

	# clear global error indicator
	$! = 0;

	$token->{code} = 500;
	$token->{diagnostic} = "";

	my $result = exch2aux($token, $data);
	
	delete $token->{code};
	delete $token->{diagnostic};

	my $name = $token->{account}->{name};
	if ($token->{updateP}) {
		my $result = $token->{sasl}->store($name, $token->{account});

		if (! $!) {
			# NO-OP
			;
		}
		elsif ($! < 0) {
			$token->{log}->entry("notify", "sasl::store", $result->{code}, ": ",
			            $result->{diagnostic});
		}
		else {
			$token->{log}->entry("notify", "sasl::store", $!);
		}
	}

	release($token, $name);

	$token->{rspP} = 0;

	return $result;
}


#
# auxiliary support for exch2
#
sub exch2aux {
	my $token = shift;
	my $data  = shift;

	if ($data->{abortP}) {
		return "<complete />";
	}

	my $response = trim(lc $data->{response});
	unless ($response) {
		$token->{diagnostic} = "<response> element poorly formed";
		my $err = "<error code='$token->{code}'>\n";
		$err .= xml_cdata($token->{diagnostic});
		$err .= "</error>";

		$! = -7;
		return $err;
	}

	my @parts = split(":", $response);

	my %otp = %{ $token->{authentication} };
	
	my $key;
	my $type = join("/", $parts[0], $#parts+1);
	if ($type eq "hex/2" || $type eq "init-hex/4") {
		$key = parse_hex($token, $parts[1]);
	}
	elsif ($type eq "word/2" || $type eq "init-word/4") {
		$key = parse_words($token, $parts[1]);
	}
	else {
		my $len = scalar @parts;

		if ($len) {
			my $s = ($len != 1) ? "s" : "";
			$token->{diagnostic} = "not expecting \"$parts[0]\" with $len argument${s}";
		}
		else {
			$token->{diagnostic} = "excepting a non-empty <response> element";
		}
	}

	if ($token->{diagnostic}) {
		my $err = "<error code='$token->{code}'>\n";
		$err .= xml_cdata($token->{diagnostic}) . "\n";
		$err .= "</error>";

		$! = -7;
		return $err;
	}

	if ($token->{loseP}) {
		$token->{sasl}->failure($token->{name},
		                         { code => 550, diagnostic => "account unknown" });

		$! = -7;
		return "<error code='535' />";
	}

	my $otpkey = unpack("h*", $otp{key});
	my $algkey;
	if ($otp{algorithm} eq "md5") {
		$algkey = MD5_compute($key, $otp{sequence});
	}
	elsif ($otp{algorithm} eq "sha1") {
		$algkey = SHA1_compute($key, $otp{sequence});
	}

	if ($otpkey ne $algkey) {
		$token->{sasl}->failure($token->{name},
		                         { code => 535, diagnostic => "digest mismatch" });

		$! = -7;
		return "<error code='535' />";
	}

	$token->{updateP} = 1;
	$otp{sequence}--;
	$otp{key} = pack("h*", $key);

	my %init = ();

	if ($parts[0] eq "init-hex") {
		%init = %{ parse_params($token, $otp{seed}, $parts[2]) };
		$init{key} = pack("h*", parse_hex($token, $parts[3]));
	}
	elsif ($parts[0] eq "init-word") {
		%init = %{ parse_params($token, $otp{seed}, $parts[2]) };
		$init{key} = pack("h*", parse_word($token, $parts[3]));
	}

	if ($token->{diagnostic}) {
		my $err = "<error code='$token->{code}'>\n";
		$err .= xml_cdata($token->{diagnostic}) . "\n";
		$err .= "</error>";

		$! = -7;
		return $err;
	}

	%otp = %init;

	$token->{account}->{authentication} = { %otp };

	$token->{sasl}->login($token->{name});

	return "<complete />";
}


#
# parse OTP hex phrase
#
sub parse_hex {
	my $token = shift;
	my $text  = shift;

	# clean up input hex info a little
	my $hex = $text;
	$hex =~ s/\s//g;

	if (length($hex) != 16) {
		$token->{code} = 501;
		$token->{diagnostic} = "expecting 16 hexadecimal characters";
	}

	# decode hex string
	my $result = unpack("h*", $hex);

	unless ($result) {
		$token->{code} = 501;
		$token->{diagnostic} = "invalid hexadecimal value";
		return "";
	}

	$result;
}


#
# parse OTP word phrase
#
sub parse_words {
	my $token = shift;
	my $text  = shift;

	my @words = split(/\s+/, trim($text));

	if ($#words != 5) {
		my $s = ($#words != 0) ? "s" : "";

		$token->{code} = 501;
		$token->{diagnostic} = "expecting 6, not " . $#words+1 . ", word${s}";
	}

	$text = join(" ", @words);

	my $result = Space::OTP::word_to_hex($text); ##  otp_words decode
	unless ($result) {
		$token->{code} = 501;
		$token->{diagnostic} = $!;
		return "";
	}

	$result;
}


#
# parse OTP challenge parameters
#
sub parse_params {
	my $token = shift;
	my $text  = shift;
	my $oseed = shift;

	my @terms = split(/\s+/, $text);
	if ($#terms != 2) {
		$token->{code} = 501;
		$token->{diagnostic} = "expecting 3 terms for new parameters";
		return "";
	}

	my $algorithm = $terms[0];

	# supported algorithm requested?
	if ($algorithm !~ /^(md5|sha1)$/) {
		$token->{code} = 504;
		$token->{diagnostic} = "only the MD5 and SHA1 algorithms are supported";
	}

	my $sequence = ($terms[1] =~ /(\d+)/, $1);
	if ($sequence < 100 || $sequence > 9999) {
		$token->{code} = 501;
		$token->{diagnostic} = "sequence-number not in 100..9999 range";
		return "";
	}

	my $seed = $terms[2];
	if (length($seed) < 1 || length($seed) > 16) {
		$token->{code} = 501;
		$token->{diagnostic} = "seed length not in 1..16 range";
		return "";
	}

	if ($seed !~ /^[A-Za-z0-9]+$/) {
		$token->{code} = 501;
		$token->{diagnostic} = "seed is poorly formed";
	}

	if ($seed ne $oseed) {
		$token->{code} = 534;
		$token->{diagnostic} = "seed not updated";
	}

	return { algorithm => $algorithm, sequence => $sequence, seed => lc($seed) };
}


#
# library version info
#
sub version {
	return __PACKAGE__ . " " . $VERSION . "r" . $RELEASE;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

