#
#  file: Frame.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: BXXP framing protocol implementation
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::BXXP::Frame;

# version of perl we expect
require 5.005;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw( FRAME_ERROR
                 INITIAL_SND_WINDOW_SIZE INITIAL_RCV_WINDOW_SIZE
                 MAX_CHANNEL_NUMBER MAX_PAYLOAD_SIZE MAX_SEQUENCE_NUMBER
                 MAX_SERIAL_NUMBER
               );


# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/BXXP/Frame.pm#13 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use


# spacekit logging support
use Space::Log;

# spacekit general utility support
use Space::Util qw(trimleft trimright);


# some useful constants
use constant FRAME_READ_OK           =>             1;
use constant FRAME_READ_TIMEOUT      =>            -1;
use constant FRAME_READ_ERROR        =>            -2;
use constant FRAME_READ_EOF          =>            -3;
use constant FRAME_ERROR             =>            -7;

use constant INITIAL_SND_WINDOW_SIZE =>         4_096;
use constant INITIAL_RCV_WINDOW_SIZE =>        65_407;

use constant MAX_CHANNEL_NUMBER      =>           255;
use constant MAX_PAYLOAD_SIZE        => 2_147_483_647;
use constant MAX_SEQUENCE_NUMBER     => 4_294_967_295;
use constant MAX_SERIAL_NUMBER       =>        32_767;

# read/write buffer size
use constant BUFSIZ                  =>        65_536;

# <CR><LF> sequence for network I/O
use constant CRLF                    => "\015\012";


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;


#
# instantiate a new frame object
#
sub new {
	my $this = shift;
	my $type = ref($this) || $this;

	my $self = {};
	bless $self, $type;

	# major.minor version info
	$self->{_VERSION} = $VERSION;
	# release (patch) level
	$self->{_RELEASE} = $RELEASE;

	$self;
}


#
# state variables
#
# log:      object for logging package
# inputC:   channel for reading
# outputC:  channel for writing
# readable: semaphore
# buffer:   buffer to read
# serial:   last serial-number used
# rcvhdr:   serialized array of key/value pairs
# rcvbuf:   fragmented packet
# rcvque:   queue of received frames
# serials:  serialized array mapping serial-numbers to channels
# seqvars:  serialized array of sequence space information
#


#
# initialise frame info
#
sub init {
	my $self = shift;

	# we require a 32-bit architecture
	my $bits = 0;
	my $val  = 1;
	while ($val > 0) {
		$bits++;
		$val <<= 1;
	}
	unless ($bits == 32) {
		$! = FRAME_ERROR;
		return { code => 451, diagnostic =>  "expecting a 32-bit architecture..." };
	}

	# logging token
	$self->{log} = shift;

	# input channel (filehandle)
	$self->{inputC} = shift;

	# output channel (filehandle)
	$self->{outputC} = shift || "";
	# map output to input if none supplied
	$self->{outputC} = $self->{inputC}
		unless ($self->{outputC});

	my $oldfh = select $self->{outputC}; $| = 1; select $oldfh;

	# am i a listener?  effects initial action on channel 0
	my $listenerP = shift || 0;

	# am i a listener or initiator (server or client)?
	if ($listenerP) { 
		# {serial#,remote} = channel
		$self->{serials}->{"0,remote"} = 0;
	}
	else {
		# {serial#,local} = channel
		$self->{serials}->{"0,local"} = 0;
	}

	$self->{buffer}   = "";
	$self->{rcvbuf}   = ();
	$self->{rcvhdr}   = ();
	$self->{rcvque}   = ();
	$self->{readable} = 0;
	$self->{seqvars}  = ();
	$self->{serial}   = 0;

	$self->{_VERSION} = $VERSION;

	# here we go
	$self->start(0);

	if ($debug > 2) {
		$self->_dump();
	}

	0;
}
	

#
# finalise frame session
#
sub fin {
	my $self = shift;

	$self = {};

	0;
}


#
# start frame session
#
sub start {
	my $self = shift;

	my $channel = shift;

	# set default frame parameters -- tunables
	$self->{seqvars}->[$channel]->{'rcv.nxt'} = 0;
	$self->{seqvars}->[$channel]->{'rcv.wnd'} = INITIAL_RCV_WINDOW_SIZE;
	$self->{seqvars}->[$channel]->{'snd.una'} = 0;
	$self->{seqvars}->[$channel]->{'snd.nxt'} = 0;
	$self->{seqvars}->[$channel]->{'snd.wnd'} = INITIAL_SND_WINDOW_SIZE;

	0;
}


#
# frame failure handler
#
sub fail {
	my $self = shift;

	my $channel = shift;

	# remove sequence space info
	undef $self->{seqvars}->[$channel];

	# remove all references to this channel
	my ($k, $v);
	while (($k, $v) = each %{$self->{serials}}) {
		if ($channel == $v) {
			delete $self->{serials}->{$k};
		}
	}

	\$self->{serials};
}


#
# frame reader
#
sub read {
	my $self = shift;

	my $timeout = shift || undef;
	my $block   = shift || 0;

	# 
	if (!$block && ($#{$self->{rcvque}} >= 0)) {
		my $hdr = shift @{$self->{rcvque}};
		return $hdr;
	}

	# what are we waiting for?
	my $waitingFor = 'message';

	my ($ackno, $channel, $seqno, $serial, $size, $status, $window);
	my ($keyword, $line, $more, $payload, $tags, $x);
	my %header = ();

	# main processing
	while (1) {
		# input loop
		while (1) {
			if ($waitingFor =~ /^(message|tags|trailer)$/i) {
				# do we have a complete line?
				if (($x = index($self->{buffer}, "\n")) > -1) {
					last;
				}
			}
			elsif ($waitingFor =~ /^(payload)$/i) {
				# get as much payload as we can
				if (($x = length($self->{buffer})) > 0) {
					last;
				}
			}

			$self->_readable($timeout);

			# read error or timeout?
			if ($self->{readable} == FRAME_READ_TIMEOUT) {
				print STDERR "<<< zzz\n"
					if ($debug > 2);

				$! = FRAME_READ_TIMEOUT;
				return { code => 451, diagnostic => "timeout" }; 
			}
			elsif ($self->{readable} == FRAME_READ_ERROR) {
				print STDERR "<<< err\n"
					if ($debug > 2);

				$! = FRAME_READ_ERROR;
				return { code => 451, diagnostic => "network error" };
			}
			elsif ($self->{readable} == FRAME_READ_EOF) {
				print STDERR "<<< eof\n"
					if ($debug > 2);

				$! = FRAME_READ_EOF;
				return { code => 451, diagnostic => "premature end-of-file" };
			}
			
			$self->{readable} = 0;
		} ## input loop

		# process the data we just read
		if ($waitingFor =~ /^(message|tags|trailer)$/i) {
			# grab a line
			$line = trimright(substr($self->{buffer}, 0, $x));
			$self->{buffer} = substr($self->{buffer}, $x+1);

			print STDERR "<<< $line\n"
				if ($debug > 2);
		}
		elsif ($waitingFor =~ /^(payload)$/i) {
			# where is $size first set???
			if ($size < $x) {
				$x = $size;
			}

			$payload .= substr($self->{buffer}, 0, $x);
			$self->{buffer} = substr($self->{buffer}, $x);

			print STDERR "<<< read $x octets"
				if ($debug > 2);
			print STDERR " ($payload)"
				if ($debug > 3);
			print STDERR "\n"
				if ($debug > 2);
			
		}

		if ($waitingFor =~ /^(message)$/i) {
			# NO-OP
			;
		}
		elsif ($waitingFor =~ /^(tags)$/i) {
			$tags .= $line . "\n";

			if (! length($line)) {
				$header{tags} = trimright($tags);
			}

			if ($size > 0) {
				$payload = "";
				$waitingFor = 'payload';
			}
			else {
				$header{data} = "";
				$waitingFor = 'trailer';
			}

			next;
		}
		elsif ($waitingFor =~ /^(payload)$/i) {
			$size -= $x;
			unless ($size) {
				$header{data} = $payload;
				$waitingFor = 'trailer';
			}

			next;
		}
		elsif ($waitingFor =~ /^(trailer)$/i) {
			if ($line ne "END") {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic =>  "expecting trailer, not $line" };
			}

			$self->SEQ($channel);

			if (!$more) {
				if ($block) {
					push(@{$self->{rcvque}}, { %header });
					$waitingFor = 'message';

					next;
				}

				return \%header;
			}

			# save partial message and continue
			$self->{rcvhdr}->{$serial} = { %header };
			$self->{rcvbuf}->{$serial} = $payload;

			$waitingFor = 'message';

			next;
		}

		$line = trimright($line);
		$keyword = ($line =~ /^(\w+)/, $1);

		# continuation frame?
		my $cont = 0;

		my ($diagnostic, $prev, $prevtag);
		my ($rcv_nxt, $rcv_wnd);

		if ($keyword eq "SEQ") {

			if ($line !~ /^SEQ (\d+) (\d+) (\d+)/) {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic => "invalid parameters: $line" };
			}
			else {
				($channel, $ackno, $window) = ($1, $2, $3);

				if ( ($channel < 0)                    ||
			       ($channel > MAX_CHANNEL_NUMBER)   ||
			       ($ackno < 0)                      ||
			       ($window < 0)                     ||
			       ($window > MAX_PAYLOAD_SIZE)
				   ) {
					$! = FRAME_ERROR;
					return { code => 451, diagnostic => "unacceptable parameters: $line" };
				}
			}

			$self->{seqvars}->[$channel]->{'snd.una'} = $ackno;
			$self->{seqvars}->[$channel]->{'snd.wnd'} = $window;

			$self->{log}->entry("debug", "SND", $channel,
			            "una=$ackno", "wnd=$window",
			            "next=$self->{seqvars}->[$channel]->{'snd.nxt'}");

			if ($block) {
				return undef;
			}

			next;
		}
		elsif ($keyword eq "REQ") {

			if ($line !~ /^REQ (.) (\d+) (\d+) (\d+) (\d+)/) {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic => "invalid parameters: $line" };
			}
			else {
				($cont, $serial, $seqno, $size, $channel) = ($1, $2, $3, $4, $5);

				if ( ($serial < 0)                   ||
			       ($seqno < 0)                    ||
			       ($size < 0)                     ||
			       ($size > MAX_PAYLOAD_SIZE)      ||
			       ($channel > MAX_CHANNEL_NUMBER) ||
			       ($channel < 0)
				   ) {
					$! = FRAME_ERROR;
					return { code => 451, diagnostic => "unacceptable parameters: $line" };
				}
			}

			$prev->{$prevtag = "channel"} = $channel;
			# clear previous data
			%header = ( channel => $channel );

			$self->{serials}->{"$serial,remote"} = $channel;
		}
		elsif ($keyword eq "RSP") {

			if ($line !~ /^RSP (.) (\d+) (\d+) (\d+) (.)/) {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic => "invalid parameters: $line" };
			}
			else {
				($cont, $serial, $seqno, $size, $status) = ($1, $2, $3, $4, $5);

			  if ( ($serial < 0)              ||
			       ($seqno < 0)               ||
			       ($size < 0)                ||
			       ($size > MAX_PAYLOAD_SIZE)
				   ) {
					$! = FRAME_ERROR;
					return { code => 451, diagnostic => "unacceptable parameters: $line" };
				}
			}

			if ($status !~ /^(\+|\-)$/) {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic => "invalid status indicator: $line" };
			}

			unless (exists $self->{serials}->{"$serial,local"}) {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic => "REQ $serial never sent" };
			}

			$channel = $self->{serials}->{"$serial,local"};

			if ($cont eq ".") {
				delete $self->{serials}->{"$serial,local"};
			}

			$prev->{$prevtag= "status"} = $status;
			# clear previous data
			%header = (
			            channel    => $channel,
			            status     => $status,
			            diagnostic => trimleft($diagnostic),
			          );
		}
		else {
			$! = FRAME_ERROR;
			return { code => 451, diagnostic => "unacceptable keyword: $line" };
		}

		$header{keyword} = $keyword;
		$header{serial} = $serial;

		if ($cont eq ".") {
			$more = 0;
		}
		elsif ($cont eq "*") {
			$more = 1;
		}
		else {
			$! = FRAME_ERROR;
			return { code => 451, diagnostic => "unknown continuation indicator: $line" };
		}
		
		$rcv_nxt = $self->{seqvars}->[$channel]->{'rcv.nxt'};
		if ($rcv_nxt ne $seqno) {
			$! = FRAME_ERROR;
			return { code => 451, diagnostic => "expecting seqno=$rcv_nxt, got $seqno" };
		}
		$self->{seqvars}->[$channel]->{'rcv.nxt'} += $size;

		$rcv_nxt = $self->{seqvars}->[$channel]->{'rcv.nxt'};
		$rcv_wnd = $self->{seqvars}->[$channel]->{'rcv.wnd'};
		$self->{log}->entry("debug", "RCV", $channel, "nxt=$rcv_nxt",
		            "win=$rcv_wnd");

		if (exists $self->{rcvhdr}->{$serial}) {
			my $first = $self->{rcvhdr}->{$serial};

			if ($first->{$prevtag} ne $prev->{$prevtag}) {
				$! = FRAME_ERROR;
				return { code => 451, diagnostic => "$prev mismatch: $line" };
			}

			delete $self->{rcvhdr}->{$serial};

			$payload = $self->{rcvbuf}->{$serial};
			delete $self->{rcvbuf}->{$serial};

			if ($size > 0) {
				$waitingFor = 'payload';
			}
			else {
				$waitingFor = 'trailer';
			}
		}
		else {
			$waitingFor = 'tags';
			$tags = "";
		}

	} ## main loop

	0;
}


#
# SEQ message
#
sub SEQ {
	my $self = shift;
	
	my $channel = shift;

	my $s = "";
	my $dP = 0;

	# debug mode?
	if ($debug < 3) {
		$dP = 0;
	}
	elsif ($dP = (fileno($self->{outputC}) == fileno(STDOUT))) {
		$s = ">>> ";
	}

	my $fh = $self->{outputC};

	my $rcv_nxt = $self->{seqvars}->[$channel]->{'rcv.nxt'};
	my $rcv_wnd = $self->{seqvars}->[$channel]->{'rcv.wnd'};

	my $msg = "${s}SEQ $channel $rcv_nxt $rcv_wnd" . CRLF;

	print $fh $msg;
	if ($dP) {
		print STDERR ">>> $msg";
	}

	0;
}


#
# frame REQ generator
#
sub REQ {
	my $self = shift;

	my $channel = shift;
	my $payload = shift || "";

	my $serial = ++$self->{serial};

	$self->{serials}->{"$serial,local"} = $channel;

	$self->_write($channel, "REQ", $serial, $channel, $payload);

	return $serial;
}


#
# frame RSP generator
#
sub RSP {
	my $self = shift;

	my $serial = shift;
	my $status = shift;

	my $payload    = shift || "";

	my $channel = $self->{serials}->{"$serial,remote"};
	unless ($channel >= 0) {
		# fatal error
		$! = FRAME_ERROR;
		return { code => 451, diagnostic => "REQ $serial never received ($channel)" };
	}

	# remove the outstanding serial info
	delete $self->{serials}->{"$serial,remote"};

	$self->_write($channel, "RSP", $serial, $status, $payload);

	0;
}


#
# library version info
#
sub version {
	my $self = shift;

	return __PACKAGE__ . " " . $self->{_VERSION} . "r" . $self->{_RELEASE};
}


#
# internal/private methods
#

#
# frame writer
#
sub _write {
	my $self = shift;

	my $channel = shift;
	my $head    = shift;
	my $serial  = shift;
	my $tail    = shift;
	my $payload = shift;

	# clear global error indicator
	$! = 0;

	my $s = "";
	my $dP = 0;

	if ($debug < 3) {
		$dP = 0;
	}
	elsif ($dP = (fileno($self->{outputC}) == fileno(STDOUT))) {
		$s = ">>> ";
	}

	my $firstP = 1;
	my $size = length($payload);

	if ($debug > 1) {
		$self->{log}->entry("debug", "frame::_write", "channel ($channel)",
		            "head ($head)", "serial ($serial)", "tail ($tail)",
		            "payload ($payload)", "size ($size)")
	}

	my $fh = $self->{outputC};

	my ($x, $next, $seqvars, $more, $header);

	while (1) {
		$seqvars = $self->{seqvars}->[$channel];
		$x = $seqvars->{'snd.una'};
		$x += $seqvars->{'snd.wnd'};
		if (($next = $seqvars->{'snd.nxt'}) > $x) {
			# sequence wrap
			$next += -(MAX_SEQUENCE_NUMBER);
		}

		$x -= $next;

		$self->{log}->entry("debug", "SND", $channel,
		  "una=$seqvars->{'snd.una'}", "wnd=$seqvars->{'snd.wnd'}",
		  "nxt=$seqvars->{'snd.nxt'}", "avail=$x", "want=$size");

		if (! $x) {
			# timeout in seconds
			$self->read(5*60);
			next;
		}

		
		# do we need continuation frames?
		if ($size <= $x) {
			$x = $size;
			$more = ".";
		}
		else {
			$more = "*";
		}

		$header = "$head $more $serial $seqvars->{'snd.nxt'} $x $tail";
		if ($firstP) {
			$firstP = 0;
			$header .= CRLF;
		}

		print $fh $s, $header, CRLF;
		
		if ($dP) {
			print STDERR ">>> $header", ((!$firstP) ? "\n" : "");
		}

		print $fh substr($payload, 0, $x);
		if ($dP) {
			print STDERR ">>> write $x octets";
			print STDERR " (",substr($payload,0,$x),")"
				if ($dP > 1);
			print STDERR "\n";
		}

		print $fh $s, "END", CRLF;
		if ($dP) {
			print STDERR ">>> END", "\n";
		}

		$self->{seqvars}->[$channel]->{'snd.nxt'} += $x;

		$payload = substr($payload, $x);

		# end of data?
		$size -= $x;
		if ($size == 0) {
			return 0;
		}
	}
}


#
# is the I/O channel readable?
#
sub _readable {
	my $self = shift;
	my $timeout = shift;

	my $fh = $self->{inputC};
	my $n;

	if ( (! defined($n = $self->_read($fh, $timeout))) || ($n < 0) ) {
		# read error
		$self->{log}->entry("user", "frame::_readable", $!);

		$self->{readable} = FRAME_READ_ERROR;
	}
	elsif ($n == 0) {
		# TIMEOUT
		$self->{readable} = FRAME_READ_TIMEOUT;
	}
#	elsif (eof($fh)) {
#		# EOF
#		$self->{readable} = FRAME_READ_EOF;
#	}
	else {
		# OK
		$self->{readable} = FRAME_READ_OK;
	}

	if ($self->{readable} != FRAME_READ_OK) {
		$self->{log}->entry("system", $!);
	}

	0;
}


#
# my read function
#
sub _read {
	my $self = shift;

	my $fh = shift;
	my $timeout = shift;

	# is file descriptor ready to read?
	my $n = _select($fh, $timeout, 1);
	my $offset = length($self->{buffer});

	if ($n > 0) {

		# read as much as we can
READ:
		$n = sysread($fh, $self->{buffer}, BUFSIZ, $offset);

		# interrupted syscall?
		if (!defined($n)) {
			redo READ if ($! =~ /^Interrupted/);
		}

		print STDERR "DEBUG: _read: n: ($n) offset: ($offset) buf: ($self->{buffer})\n"
			if ($debug > 4);

		$offset += $n;
	}

	$n;
}


#
# select on file descriptor(s)
#
sub _select {
	my $fh = shift;
	my $timeout = shift;
	my $rw = shift;

	my ($rin, $win);

	$rin = '';
	vec($rin, fileno($fh), 1) = 1;

	$win = $rw ? undef : $rin;
	$rin = undef unless $rw;

	my $nfound = select($rin, $win, undef, $timeout);

	return $nfound;
}


#
# timeout handler
#
sub _timeout {
	my $self = shift;

	$self->{readble} = FRAME_READ_TIMEOUT;

	0;
}


#
# debug support
#
use Data::Dumper;

sub _dump {
	my $self = shift;

	print STDERR Dumper($self), "\n";

	0;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::BXXP::Frame - BXXP Framing Protocol for the Perl SpaceKit.

=head1 SYNOPSIS

C<use Space::BXXP::Frame;>

=head1 DESCRIPTION

Provides BXXP framing protocol implementation.  There should be no need for
most applications to use the framing protocol directly.

Please see http://xml.resource.org/profiles/BXXP/bxxp.html for more information
on BXXP.

=head1 METHODS

=over 4

=item new();

Class constructor, returns a new C<Space::BXXP::Frame> object on success, an
undefined value otherwise.  Optionally takes the same parameters as the init()
routine to create and initialize a frame object in one step.

=item init($log, $inputHandle, $outputHandle);

Initialize frame info.  The optional outputHandle defaults to the inputHandle.
$log is a log token from a successful call to C<Space::Log::log_init> which
allows informational logging support from the frame methods.

=item fin();

Finalizes an instance of the API, uninitializing the object.

=item start($channel);

Initialize frame session on the specified channel.

=item fail($channel);

Cleans up state information for the specified channel.  Should be called on
frame error.

=item read($timeout, $block);

Read a complete frame message, I<block>ing if appropriate, or until I<timeout>
occurs.

=item SEQ($channel);

Generates a SEQ message on the specified channel.

=item REQ($channel, $payload);

Generates a REQ message on the specified channel, sending the optional payload
as content of the message.

=item RSP($serial, $status, $payload);

Generates a RSP message for the specified message serial number with the
provided status.  payload is optional additional message content.

=item version();

Library version info.

=back

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

