#
#  file: Log.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: common logging support
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::Log;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw();

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/Log.pm#12 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# date formatting support ala strftime(3C)
use Date::Format;

use Errno qw(:POSIX EEXIST EINVAL);
use Fcntl;
use FileHandle;

# MIME mail support in _mail()
use MIME::Lite;


# spacekit general utility support
use Space::Util qw(trim);


# other global vars
use vars qw($debug);

# debug mode?
$debug = 0;


#
# state variables
#
# outputC:    filehandle for log file
# file:       file-path for log file
# ident:      service name
# originator: address to send critical events from
# recipient:  address to send critical evens to
#


#
# instantiate a new log object
#
sub new {
	my $this = shift;
	my $type = ref($this) || $this;

	my $self = {};
	bless $self, $type;

	# major.minor version info
	$self->{_VERSION} = $VERSION;
	# release (patch) level
	$self->{_RELEASE} = $RELEASE;

	# init options provided?
	if (@_) {
		$self->init(@_);
	}

	$self;
}



#
# initialise Log info
#
sub init {
	my $self = shift;

	# log file to use
	$self->{file} = shift;

	# who am i logging as?
	$self->{ident} = sprintf("%-8.8s", Space::Util::trim(shift));

	$self->{recipient}  = shift || "";
	$self->{originator} = shift || "";

	# pid to provide in log file
	$self->{logpid} = sprintf("%06d", $$ % 65535);

	# output channel (filehandle)
	$self->{outputC} = "";
}
	

#
# finalise Log object
#
sub fin {
	my $self = shift;

	# close the output channel if it exists
	if ($self->{outputC}) {
		close $self->{outputC};
	}

	0;
}


#
# generate Log entry
#
sub entry {
	my $self = shift;

	# severity level to log at
	my $level = shift;
	# arguments/messages to log
	my @args  = @_;

	# clear global error indicator
	$! = 0;

	# don't log if we are not interested in debug messages OR
	# our debug level is lower than that being logged
	my $x;
	$x = $level =~ /debug(\d+)/ && $1;
	if ( ( $level =~ /^debug/ && ! $debug) || ($x > $debug) ) {
		return 0;
	}

	# current time (in seconds)
	my $now = time;

	# open our log file if we don't have one already
	if (! $self->{outputC}) {
		# attempt to open log file
		my $fh = new FileHandle $self->{file}, O_WRONLY|O_CREAT|O_APPEND;

		return 0 unless (defined($fh));

		# save the file handle
		$self->{outputC} = $fh;
	}

	# clean up log message a little
	my $msg = join(" ", @args);
	$msg =~ s/\r?\n/ /gs;
	# remove any remaining CR
	$msg =~ s/\r//gs;

	if (length($msg) > 200 && ! $debug) {
		$msg = substr($msg, 0, 197) . "...";
	}

	# log message prefix
	my $prefix = join(" ", time2str("%m/%d %T", $now), $self->{ident},
	                   $self->{logpid});

	# write the log message
	eval {
		my $fh = $self->{outputC};
		$fh->autoflush(1);
		print $fh sprintf("%s %-7.7s %s\n", $prefix, $level, $msg);
	};

	print STDERR "debug: ", $msg, "\n"
		if ($debug > 1);

	if ($level =~ /^(error|fatal|notice|notify|system)$/i) {
		my $einfo = $!;

		$einfo =~ s/\r?\n/ /gs;
		if (length($einfo) > 200) {
			$einfo = substr($einfo, 0, 197) . "...";
		}

		if ($einfo) {
			eval {
				my $fh = $self->{outputC};
				$fh->autoflush(1);
				print $fh "$prefix trace $einfo", "\n";
			};
		}

		if ($debug > 1) {
			print STDERR "debug: ", $einfo, "\n";
		}
		else {
			my $suffix = join(" ", @args);

			use Sys::Hostname;
			my $host = eval { hostname(); };

			# default to localhost if hostname lookup fails
			$host = "localhost" if $@;

			$self->_mail("$host $level", "$prefix $level $suffix", $!, $einfo);

			if ($level eq "notice" || $level eq "notify") {
				$self->_syslog("local2", "notice", $self->{ident}, $msg);
			}
		}
	}

	0;
}


#
# generate MIME mail message
#
sub _mail {
	my $self = shift;

	my $subj = shift;
	my $text = shift;
	my $ecode = shift;
	my $einfo = shift;

	my $from;

	# no one to send to?  don't bother...
	return 0 unless ($from = $self->{recipient});

	$from = $self->{originator}
		if ($self->{originator});

	# create the MIME message
	my $mime = new MIME::Lite
	               From    => $from,
	               To      => $self->{recipient},
	               Subject => $subj,
	               Type    => "text/plain",
	               Data    => $text;

	# send it
	$mime->send();

	0;
}


# syslog(3) support in _syslog()
use Sys::Syslog;


#
# system syslog for certain message types
#
sub _syslog {
	my $self = shift;

	my $facility = shift;
	my $priority = shift;
	my $tag      = shift;
	my $msg      = shift;

	# open syslog channel
	openlog($tag, "cons,pid,nowait", $facility);

	syslog($priority, $msg);

	# down with channel
	closelog();

	0;
}


#
# library version info
#
sub version {
	my $self = shift;

	return $self->{_VERSION} . "r" . $self->{_RELEASE};
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::Log - Library of logging functions for the Perl SpaceKit

=head1 SYNOPSIS

 use Space::Log;

 my $log = Space::Log->new($file, $ident, $recipient, $originator);

 $log->entry("error", "oops: ", $!)
	if ($!);

 $log->fin();

=head1 DESCRIPTION

Space::Log provides common logging capabilities for SpaceKit.

=head1 FUNCTIONS

=over 4

=item new()

This is a class method, the contructor for Space::Log.  This method
optionally accepts the same parameters as init() to perform both
operations in a single step.

=item init($file, $ident, $recipient, $originator)

Initializes an instance of the API.  The parameters are:

    file:
        the name of the log file;
    ident:
        the caller's logging identity (e.g., "space");
    recipient:
        if present, the address to send critical messages to; and,
    originator:
        if present, the address to send critical messages from.

=item entry

Appends an entry to the log file.  The parameters are:

    level:
        a textual string that defines the form of the "args"
        parameter; one of:

        debugN
            function args... (where N > 0, and defaults to 1)
        error
            function result
        notify
            function result
        fatal
            errorInfo result
        info
            begin/end args...
        stats
            name value...
        system
            result
        user
            result

    args:
        a list of the textual portion of the log entry.

=item fin

Finalizes an instance of the log package, uninitializing the object.


=item version

Returns the current library version.

=back

=head1 EXAMPLES

  use Space::Log;

  my $file = "logs/test.log";
  my $ident = "myident";
  my $originator = "user\@other.domain.com";
  my $recipient = "user\@some.domain.com";

  my $log = Space::Log->new($file, $ident, $recipient, $originator);

  $log->entry("info", "begin", "starting tests");

  my $status = run_some_tests();

	if ($status && $!) {
		$log->entry("error", "run_some_tests", $!);
	}

  $log->entry("info", "end", $status);

  $log->fin();

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

