#
#  file: Mixer.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: Blocks mixer (client) API
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::Mixer;

# version of perl we expect
require 5.005;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw();

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/Mixer.pm#19 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# network socket I/O support
use IO::Socket;

# MIME Base64 encode/decode
use MIME::Base64;

# MIME parsing/processing support
use MIME::Parser;

# generic XML parsing support
use XML::Parser::Expat;


# spacekit BXXP framing protocol support
use Space::BXXP::Frame;

# spacekit logging support
use Space::Log qw();

# spacekit OTP support
use Space::OTP qw(compute_otp);

# spacekit general utility support
use Space::Util qw(timer_log trim);

# spacekit XML support
use Space::XML qw(xml_cdata xml_escape_av xml_error xml_normalize xml_parse);


# some useful constants

use constant BXXP_PORT	=> 10288;


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;



#
# instantiate a new mixer object
#
sub new {
	my $this = shift;
	my $type = ref($this) || $this;

	my $self = {};
	bless $self, $type;

	# TLS package initialised?
	$self->{_tlsP} = 0;

	# array of blocks
	$self->{_cache} = ();

	# major.minor version info
	$self->{_VERSION} = $VERSION;
	# release (patch) level
	$self->{_RELEASE} = $RELEASE;

	# init options provided?
	if (@_) {
		$self->init(@_);
	}

	$self;
}


#
# namespace variables
#
# _cache:    array of blocks
# _tlsP:     TLS package initialised?
#
# state variables
#
# log:       object for logging package
# server:    domain name of offered service
# options:   options to mixer::init
# socket:    channel for reading
# frame:     BXXP::Frame object
# block:     token for datastore package
# reqno:     last request number used
# eofP:      responded to request to release connection
# errP:      saw error on connection
# notifyA:   array of notification vectors
# debugZ:    initial debug setting
# faststart: doing fast start
# features:  list of features advertised in greeting
# localize:  list of localize values advertised in greeting
# uris:      list of URIs advertised in greeting
# user_channel:
#            last user-specified channel number used
#
# present only during XML parsing
#
# actualNum:  the actualNum attribute for the answers section
# additional: list of names in the additional section
# answers:    list of names in the answers section
# code:       error code
# deletions:  list of names in the deletions section
# diagnostic: error diagnostic
# envelope:   <response> or <notify> attributes
# reqStamp:   opaque string from server
# rspno:      request-number
# which:      container of interest
# uris:       list of URIs in greeting
#


#
# initialise mixer info
#
sub init {
	my $self = shift;

	# logging token
	$self->{log} = shift;

	# server host/domain
	$self->{server} = shift;

	# remaining options in hashref
	my $params = shift;

	# clear global error indicator
	$! = 0;

	# set default options
	my %options = (
		-bulk       => "false",
		-debug      => "false",
		-local      => "false",
		-mechanism  => "none",
		-port       => BXXP_PORT,
		-privacy    => "none",
		-servername => $self->{server},
	);

	# update options
	%options = (%options, %{ $params });

	$options{-generic} = Space::Util::boolean($options{-generic});
	if ($options{-generic}) {
		$options{-bulk} = "false";
		$options{-local} = "false";
	}

	$options{-bulk} = Space::Util::boolean($options{-bulk});
	if ($options{-bulk}) {
		# bulk option implies local access
		$options{-local} = "true";
	}

	$options{-debug} = Space::Util::boolean($options{-debug});
	$debug = $options{-debug}
		if ($options{-debug});

	$options{-local} = Space::Util::boolean($options{-local});

	# error indicator
	my $ecode = 0;
	my $ewhat = "";

	my ($blockT, $frameT, $socket);

	if ($options{-local}) {
		$socket = undef;
		$frameT = undef;

		# local access currently unsupported
		$blockT = undef;
	}

	$socket = IO::Socket::INET->new("$self->{server}:$options{-port}");
	if (! $socket) {
		# socket error

		$ecode = 451;
		$ewhat = "IO::Socket";
	}
	elsif (! ($frameT = Space::BXXP::Frame->new()) ||
	         ($frameT->init($self->{log}, $socket)) ) {
		# frame initialisation failed

		$ecode = 451;
		$ewhat = "Frame::init";
	}
	else {
		$blockT = undef;
	}

	if ($ecode) {
		my $err = "$ewhat: $!";

		$self->{log}->entry("error", $ewhat, $!);

		$! = -7;
		return { code => $ecode, diagnostic => $err };
	}

	# initialise some things
	$self->{block}     = $blockT;
	$self->{debugZ}    = $debug;
	$self->{eofP}      = 0;
	$self->{errP}      = 0;
	$self->{faststart} = 0;
	$self->{frame}     = $frameT;
	$self->{notifyA}   = undef;
	$self->{options}   = { %options };
	$self->{port}      = $options{-port};
	$self->{socket}    = $socket;
	$self->{reqno}     = 0;

	$self->{uris}      = ();
	$self->{user_channel} = 3;

	$self->_dump()
		if ($debug > 1);

	my $result;
	if (! $options{-local} &&
	    ( ($result = $self->bxxp_greeting()) ||
	      ($result = $self->tls_start())  ||
	      ($result = $self->sasl_start()) ||
	      (! $options{-generic} && ($result = $self->sep_start()))
	    )
	   )
	{
		if ($! == -7) {
			$self->{log}->entry("user", "$result->{code}: $result->{diagnostic}");
		}
		else {
			$self->{log}->entry("error", "mixer::bxxp_greeting",
			            "$result->{code}: $result->{diagnostic}");
		}

		if ($self->fin()) {
			$self->{log}->entry("error", "mixer::fin",
			            "$result->{code}: $result->{diagnostic}");
		}

		$! = -7;
		return $result;
	}

	0;
}
	

#
# finalise frame session
#
sub fin {
	my $self = shift;

	my $debug = $self->{debugZ};

	my $result;

	if ($self->{options}->{-local}) {
		# currently not supported
		;
	}
	else {
		if (! $self->{errP} &&
		    ! $self->{eofP} &&
		    ($result = $self->bxxp_end()) )
		{
			$self->{log}->entry("error", "mixer::bxxp_end",
			            "$result->{code}: $result->{diagnostic}");
		}
		
		if ($result = $self->{frame}->fin()) {
			$self->{log}->entry("error", "frame::fin",
			            "$result->{code}: $result->{diagnostic}");
		}

		# close the socket
		$self->{socket} = undef;
	}

	# restore global debug setting
	if (defined($::debug)) {
		$::debug = $debug;
	}

	0;
}


#
# simple exchange profile (SEP) API
#

#
# SEP fetch
#
sub fetch {
	my $self = shift;

	# name of block to fetch
	my $name = shift;

	# name of top-most element we care about
	my $top  = shift || undef;

	# clear global error indicator
	$! = 0;

	# return cached block, if available
	if (exists $self->{_cache}->{$name}) {
		return $self->{_cache}->{$name};
	}

	my $result;
	if ($self->{options}->{-local}) {
		# not currently supported
		;
	}
	else {
		my $union = "";

		$union .= "<union><intersect>\n";
		$union .= "  <compare subtree=";
		$union .= xml_escape_av($name);
		$union .= ">";
		
		if ($#{$top} >= 0) {
			$union .= "    <path attribute='name'>\n";
			$union .= "      <element property=";
			$union .= xml_escape_av($top->[0]) . " />\n";
			$union .= "</path>";
		}
		else {
			$union .= "    <path attribute='name' />";
		}

		$union .= "    <value>$name</value>\n";
		$union .= "  </compare>\n";
		$union .= "</intersect></union>";

		# perform SEP search
		$result = $self->search($union, { maxNum => 1 });
	}

	# did search succeed?
	if (! exists($result->{code}) && exists($self->{_cache}->{$name})) {
		$! = 0;
		return $self->{_cache}->{$name};
	}

	# error return
	if (! exists $result->{code}) {
		$result = { code => 450, diagnostic => "$name doesn't exist" };
	}
	$! = -7;

	return $result;
}


#
# SEP search
#
sub search {
	my $self = shift;

	my $payload = shift;
	my $opts    = shift || undef;

	# clear global error indicator
	$! = 0;

	my $reqno = ++$self->{reqno};

	if ($self->{options}->{-local}) {
		$! = -7;
		return { code => 450, diagnostic => "-local not allowed for searching" };
	}

	# search attributes
	my %attrs = (
		maxNum       => 1024,
		notification => "",
		offset       => 0,
		prevStamp    => "",
		related      => "",
	);

	# merge in supplied options
	%attrs = (%attrs, %{ $opts });

	if ($attrs{notification} && $attrs{notification} ne "false") {
		$self->{notifyA}->{reqno} = $attrs{notification};
		$attrs{notification} = "true";
	}
	else {
		$attrs{notification} = "false";
	}

	my $request = "<request reqno='$reqno'>\n";
	$request .= "  <fetch";

	while (my ($k, $v) = each %attrs) {
		$request .= " $k=" . xml_escape_av($v);
	}

	$request .= ">$payload\n";
	$request .= "  </fetch>\n";
	$request .= "</request>";

	my $frame = $self->reqrsp(1, $request);

	if ($frame->{status} ne "+") {
		$! = -7;
		return $self->exception($frame);
	}

	return $self->sep_response($reqno, $frame);
}


#
# SEP store
#
sub store {
	my $self = shift;

	# list of blocks to store
	my $blockL = shift;

	# store action to perform, default is write
	my $action = shift || "write";

	# block serial number, default is 1
	my $serial = shift || 1;

	# clear global error indicator
	$! = 0;

	my $reqno = ++$self->{reqno};

	my $text = "";
	if (! $self->{options}->{-local}) {
		$text .= "  $blockL->{body}\n";

		my $request = "<request reqno='$reqno'>\n";
		$request .= "  <store action=" . xml_escape_av($action) . ">\n";
		$request .= "$text</store>\n";
		$request .= "</request>";

		my $frame = $self->reqrsp(1, $request);

		if ($frame->{status} ne "+") {
			$! = -7;
			return $self->exception($frame);
		}

		return $self->sep_response($reqno, $frame);
	}

	# local store not currently supported

	0;
}


#
# SEP lock
#
sub lock {
	my $self = shift;

	my $subtree = shift;

	# clear global error indicator
	$! = 0;

	my $reqno = ++$self->{reqno};

	if (! $self->{options}->{-local}) {
		my $request = "<request reqno='$reqno'>\n";
		$request .= "  <lock subtree=" . xml_escape_av($subtree) . "/>\n";
		$request .= "</request>";

		my $frame = $self->reqrsp(1, $request);

		if ($frame->{status} ne "+") {
			$! = -7;
			return $self->exception($frame);
		}

		return $self->sep_response($reqno, $frame);
	}

	# local lock not currently supported

	0;
}


#
# SEP release
#
sub release {
	my $self = shift;

	# request number to release
	my $prevno = shift;

	# commit changes before release?
	my $commit = shift || 0;

	# clear global error indicator
	$! = 0;

	my $reqno = ++$self->{reqno};

	if (! $self->{options}->{-local}) {
		$commit = ($commit) ? "commit" : "rollback";

		my $request = "<request reqno='$reqno'>\n";
		$request .= "  <release prevno='$prevno' commit='$commit' />\n";
		$request .= "</request>";

		my $frame = $self->reqrsp(1, $request);

		if ($frame->{status} ne "+") {
			$! = -7;
			return $self->exception($frame);
		}

		return $self->sep_response($reqno, $frame);
	}

	# local release not currently supported

	0;
}


#
# direct BXXP access (no SEP)
#

#
# get list of available profiles
#
sub profiles {
	my $self = shift;

	return $self->{uris};
}


#
# create a generic channel to BXXP server
#
sub create {
	my $self = shift;

	my $uri = shift;
	my $data = shift || "";

	# clear global error indicator
	$! = 0;

	my $channel = $self->{user_channel} += 2;

	$self->{frame}->start($channel);

	my $start = "<start number=" . xml_escape_av($channel) . ">\n";
	$start .= "  <profile uri=" . xml_escape_av($uri) . ">";
	$start .= $data;
	$start .= "  </profile>\n";
	$start .= "</start>";

	my $frame = $self->reqrsp(0, $start);

	if ($! || ($frame->{status} ne "+")) {
		$self->{frame}->fail($channel);

		$! = -7;
		return $self->exception($frame);
	}

	return $channel;
}


#
# generic BXXP exchange
#
sub exch {
	my $self = shift;

	my $channel     = shift;
	my $mime_entity = shift;

	# clear global error indicator
	$! = 0;

	if ($channel < 5) {
		croak("channel-number not assigned by mixer::create");
	}

	my $mime_hdr = $mime_entity->head();

	my $headers = "Content-Type: " . $mime_hdr->mime_attr("content-type");

	# remove processed headers
	$mime_hdr->delete("content-type");

	# additional attributes available?
	foreach my $attr (qw(charset boundary name)) {
		my $v = $mime_hdr->mime_attr("content-type.$attr");
		if ($v) {
			$headers .= qq|; $attr="$v"|;

			# remove processed headers
			$mime_hdr->delete($attr);
		}
	}

	$headers .= "\n";

	# get any remaining headers
	$headers .= $mime_hdr->stringify();

	my $serial = $self->{frame}->REQ($channel, $mime_entity->bodyhandle()->as_string(), $headers);
	my $frame = $self->rsp($serial);

	if ($frame->{status} ne "+") {
		$! = -7;
	}
	else {
		$! = 0;
	}

	if (! $frame->{tags}) {
		$frame->{tags} = "Content-Type: text/xml\nContent-Transfer-Encoding: 8bit\n";
	}

	return build MIME::Entity Data => $frame->{data};
}


#
# scan <error> element
#
sub errscan {
	my $self = shift;

	my $mime_entity = shift;

	# clear global error indicator
	$! = 0;

	my $result = $self->parse($mime_entity->body());
	if ($!) {
		$! = -7;
	}
	elsif ($result->{which} ne "error") {
		$! = -7;
		$result = "top-level element is $result->{which}, not error";
	}
	else {
		$result = { code => $result->{code}, diagnostic => $result->{diagnostic} };
	}

	if ($!) {
		$self->{errP} = 1;
	}

	return $result;
}


#
# bxxp greeting
#

sub bxxp_greeting {
	my $self = shift;

	if ( !$self->{options}->{-generic} &&
	     $self->{options}->{-privacy} eq "none" &&
	     $self->{options}->{-mechanism} eq "none" ) {
		$self->{faststart} = 1;
	}

	if (! $self->{faststart}) {
		return $self->bxxp_greetingaux();
	}

	0;
}


sub bxxp_greetingaux {
	my $self = shift;

	# clear global error indicator
	$! = 0;

	my $frame = $self->rsp(0);

	if ($frame->{status} ne "+") {
		$! = -7;
		return $self->exception($frame);
	}

	# parse <greeting> element
	my $greeting = $self->greeting_parse($frame->{data});

	# get list of profile URIs from server <greeting>
	$self->{uris} = $greeting->{uris};

	# get list of features from server <greeting>
	$self->{features} = $greeting->{features};

	# get list of localize values from server <greeting>
	($self->{localize} = $greeting->{localize}) =~ tr [A-Z] [a-z];

	$self->{log}->entry("debug", "mixer::bxxp_greeting", $self->{features},
		$self->{localize});

	# send client greeting
	$self->{frame}->RSP(0, "+", "<greeting />");

	0;
}


#
# generate server name attribute for <start> element
#
sub bxxp_server {
	my $self = shift;

	my $server = $self->{options}->{-servername};
	if ($server) {
		$server = "serverName=" . xml_escape_av($server);
		$self->{options}->{-servername} = "";
	}

	return $server;
}


#
# bxxp session ending
#
sub bxxp_end {
	my $self = shift;

	# clear global error indicator
	$! = 0;

	my $serial = $self->{frame}->REQ(0, "");

	if ( $self->{options}->{-privacy} eq "none" &&
	     $self->{options}->{-mechanism} eq "none" ) {
		return 0;
	}

	return $self->rsp($serial, 1);
}


#
# TLS privacy (currently not supported)
#

sub tls_start {
	my $self = shift;

	# clear global error indicator
	$! = 0;

	my $privacy = $self->{options}->{-privacy};

	if ($privacy eq "none") {
		return 0;
	}
	elsif ( $privacy eq "weak" ||
	        $privacy eq "strong" ) {
		$! = -7;
		return { code => 451, diagnostic => "TLS currently not implemented" };
	}
	else {
		$! = -7;
		return { code => 500, diagnostic => "unknown privacy level: $privacy" };
	}

	# NOT-REACHED
	0;
}


#
# SASL login
#

sub sasl_start {
	my $self = shift;

	# clear global error indicator
	$! = 0;

	# DTD identities
	my $URI_SASL_ANON = "http://xml.resource.org/profiles/sasl/ANONYMOUS";
	my $URI_SASL_OTP  = "http://xml.resource.org/profiles/sasl/OTP";

	my $channel = 0;
	my $data = "";

	my $mechanism = $self->{options}->{-mechanism};

	if ($mechanism eq "none") {
		return 0;
	}
	elsif ($mechanism eq "anonymous") {
		my $trace = $self->{options}->{-trace};
		if (! $trace) {
			$! = -7;
			return { code => 500, diagnostic => "missing -trace option" };
		}

		$channel = 0;
		$data = "<start number='3'>\n";
		$data .= "  <profile uri='$URI_SASL_ANON'>\n";
		$data .= "    <blob>" . encode_base64($trace) . "</blob>\n";
		$data .= "  </profile>\n";
		$data .= "</start>";

	}
	elsif ($mechanism eq "otp") {

		my $name;
		if (! ($name = $self->{options}->{-username}) ) {
			$! = -7;
			return { code => 500, diagnostic => "missing -username option" };
		}

		my $passphrase = undef;
		my $upcall = undef;
		if ($passphrase = $self->{options}->{-passphrase}) {
			# NO-OP
			;
		}
		elsif ($upcall = $self->{options}->{-passback}) {
			# NO-OP
			;
		}
		else {
			$! = -7;
			return { code => 500, diagnostic => "missing -pass* options" };
		}

		$self->{frame}->start(3);

		$name = "\0$name";
		my $request = "<start number='3'>\n";
		$request .= "  <profile uri='$URI_SASL_OTP'>\n";
		$request .= "    <blob>" . encode_base64($name) . "</blob>\n";
		$request .= "  </profile>\n";
		$request .= "</start>";

		my $frame = $self->reqrsp(0, $request);

		if ($frame->{status} ne "+") {
			$self->{frame}->fail(3);

			$! = -7;
			return $self->exception($frame);
		}

		$data = $frame->{data};

		$data =~ s/\n//gm;
		my $x0 = index($data, ">");
		my $x1 = -1;
		$x1 = rindex($data, "</");

		if ($x0 < 0 || $x1 < 0 || $x0 > $x1) {
			$! = -7;
			return { code => 500, diagnostic => "syntax error while starting SASL" };
		}

		$data = substr($data, $x0+1, $x1-$x0-1);

		if ($data =~ /<error/gm) {
			$! = -7;
			return $self->exception($frame);
		}

		$x0 = index($data, ">");
		$x1 = rindex($data, "<");

		if ( ($data !~ /<blob/gm) ||
		     ($x0 < 0 || $x1 < 0 || $x0 > $x1) ) {
			$! = -7;
			return { code => 500, diagnostic => "syntax error while starting OTP" };
		}

		$data = Space::Util::trim(substr($data, $x0+1, $x1-$x0-1));

		eval { $data = decode_base64($data); };
		if ($@) {
			$! = -7;
			return { code => 500, diagnostic => "invalid blob encoding: $data" };
		}

		my @items = split(/\s+/, $data);
		my $seqno;
		($seqno = $items[1]) =~ /(\d+)/;
		if ( ($#items < 3) ||
		     (! $seqno)    ||
		     ($items[3] !~ /ext/) ) {
			$! = -7;
			return { code => 500, diagnostic => "invalid OTP terms: $data" };
		}

		my $challenge = join(" ", @items);
		unless ($challenge =~ /^otp-(md5|sha1) /) {
			$! = -7;
			return { code => 504, diagnostic => "unsupported OTP algorithm: otp-$1" };
		}

		my $result;
		if ($passphrase) {
			# create extended 6-word OTP value
			$result = compute_otp($challenge, $passphrase, "word");
		}
		else {
			$result = $upcall->($self->{server}, $name, $challenge);
		}

		$channel = 3;

		if ($result) {
			$data = "<blob>" . encode_base64($result) . "</blob>";
		}
		else {
			$data = "<blob status='abort' />";
		}

	}
	else {
		$! = -7;
		return { code => 500, diagnostic => "unknown mechanism: $mechanism" };
	}

	if ($channel == 0) {
		$self->{frame}->start(3);
	}

	my $frame = $self->reqrsp($channel, $data);

	if ($frame->{status} ne "+") {
		if ($channel == 0) {
			$self->{frame}->fail(3);
		}

		$! = -7;
		return $self->exception($frame);
	}

	if ($frame->{data} !~ /complete/gm) {
		$! = -7;
		return { code => 500, diagnostic => "syntax error while completing SASL" };
	}

	my $text = $frame->{diagnostic};
	$text = "successful on channel 3"
		unless ($text);
		
	$self->{log}->entry("debug", "mixer::sasl_start", "SASL: $text");

	0;
}


#
# SEP, cont'd
#

sub sep_start {
	my $self = shift;

	# SEP profile signature/identity
	my $URI_SEP = "http://xml.resource.org/profiles/SEP";

	$self->{frame}->start(1);

	my $start = "<start number='1'>\n";
	$start .= "  <profile uri='$URI_SEP' />\n";
	$start .= "</start>";

	$self->{sepno} = $self->{frame}->REQ(0, $start);

	if (! $self->{faststart}) {
		return $self->sep_startaux();
	}

	0;
}


sub sep_startaux {
	my $self = shift;

	my $frame = $self->rsp($self->{sepno});

	if ($frame->{status} ne "+") {
		$self->{frame}->fail(1);

		$! = -7;
		return $self->exception($frame);
	}

	my $text = $frame->{diagnostic};
	if (! $text) {
		$text = "started on channel 1";
	}

	$self->{log}->entry("debug", "mixer::sep_start", "SEP: $text");

	0;
}


sub sep_request {
	my $self = shift;

	my $frame = shift;

	# clear global error indicator
	$! = 0;

	my $result;

	my $parse = $self->parse($frame->{data}, 1);
	# parse problem?
	if ($!) {
		$! = -7;
		$result = { code => 500,
		            diagnostic => "$parse->{code}: $parse->{diagnostic}" };
		$self->{errP} = 1;
	}

	my %response;
	if (! $!) {
		my %attrs = ( serial => 1 );
		(%attrs) = (%attrs, %{ $parse->{envelope} } );
		my $upcall = $self->{notifyA}->{$attrs{prevno}};

		if (! $upcall) {
			$! = -7;
			$result = { code => 553, diagnostic => "persistent fetch not in progress" };
		}
		else {
			$response{reqno} = $parse->{rspno};
			$response{prevno} = $attrs{prevno};

			my $addP = 1;
			foreach my $which (qw(answers additional deletions)) {
				if ($which eq "deletions") {
					$addP = 0;
				}
				else {
					$addP = 1;
				}

				my @names;

				foreach my $body (@{ $parse->{$which} }) {
					my $name = $self->sep_update($body, $attrs{serial}, $addP);
					push(@names, $name)
						if ($name);
				}

				$response{$which} = [ @names ];
			}
		}

		$result = $upcall->(\%response);
		if ($! == 0) {
			$result = \%response;
		}
		elsif ($! < 0) {
			# NO-OP
			;
		}
		else {
			$self->{log}->entry("error", $upcall,
			            "$result->{code}: $result->{diagnostic}");
		}
	}

	return $result;
}


sub sep_response {
	my $self = shift;

	my $reqno = shift;
	my $frame = shift;

	# clear global error indicator
	$! = 0;

	my $ecode = 0;
	my (%response, $result);

	my $parse = $self->parse($frame->{data});
	# parse problem?
	if ($!) {
		$ecode = -7;
	}
	elsif ($parse->{which} eq "error") {
		$ecode = -7;
		$result = { code => 450, diagnostic => "status/payload mismatch" };
	}
	elsif (($response{reqno} = $parse->{rspno}) ne $reqno) {
		$ecode = -7;
		$result = { code => 450, diagnostic => "reqno mismatch" };
	}
	else {
		my %attrs = ( serial => 1 );
		# merge attributes
		%attrs = (%attrs, %{ $parse->{envelope} });

		foreach my $which (qw(answers additional)) {
			my @names;
			foreach my $body (@{ $parse->{$which} }) {
				my $name = $self->sep_update($body, $attrs{serial});
				push(@names, $name)
					if ($name);
			}

			$response{$which} = [ @names ];
		}

		$response{actualNum} = $parse->{actualNum};
		$response{reqStamp} = $parse->{reqStamp};

		$result = { %response };
	}

	if ($ecode) {
		$self->{errP} = 1;
	}

	# record possible error state
	$! = $ecode;

	return $result;
}


#
sub sep_update {
	my $self = shift;

	my $body   = shift;
	my $serial = shift;
	my $addP   = shift || 1;

	my $name = "";
	my $result = xml_parse($body, $serial);
	if ($! == 0) {
		$name = $result->{name};

		if ($addP) {
			$self->{_cache}->{$name} = $result->{body};
		}
		else {
			delete $self->{_cache}->{$name};
		}
	}
	elsif ($! < 0) {
		# internal parse error
		$self->{log}->entry("user", "$result->{code}: $result->{diagnostic}");
	}
	else {
		# system error
		$self->{log}->entry("error", "xml_parse",
		            "$result->{code}: $result->{diagnostic}");
	}

	return $name;
}


#
# general error/exception handler
#
sub exception {
	my $self = shift;

	my $frame = shift;

	# clear global error indicator
	$! = 0;

	my $ecode = 0;
	my $result;

	my $parse = $self->parse($frame->{data});
	if ($!) {
		$ecode = -7;
	}
	elsif ($parse->{which} eq "error") {
		$ecode = -7;
		$result = { code => $parse->{code}, diagnostic => $parse->{diagnostic} };
	}
	else {
		$result = { code => $parse->{code}, diagnostic => $parse->{diagnostic} };
	}

	if ($ecode) {
		$self->{errP} = 1;
	}

	# record possible error state
	$! = $ecode;

	return $result;
}


#
# XML parsing
#

my %_parse = ();

sub parse {
	my $self = shift;

	my $data = shift;
	my $reqP = shift || 0;

	my $result;

	# parser state info
	$_parse{actualNum}  = 0;
	$_parse{additional} = ();
	$_parse{answers}    = ();
	$_parse{code}       = 0;
	$_parse{deletions}  = ();
	$_parse{diagnostic} = "";
	$_parse{envelope}   = "";
	$_parse{reqP}       = $reqP;
	$_parse{reqStamp}   = "";
	$_parse{rspno}      = "";
	$_parse{which}      = "";

	# XML parser instance
	my $parser;
	eval {
		$parser  = new XML::Parser::Expat();
		$parser->setHandlers('Start' => \&_element_start,
		                     'End'   => \&_element_end,
		                     'Char'  => \&_cdata);

		$data = xml_normalize($data);
		$parser->parse($data);

		# release internal parser reference(s) -- REQUIRED
		$parser->release();
	};

	if ($@) {
		$self->{log}->entry("notify", "mixer::parse", $@);
		xml_error($self->{log}, "mixer", $data);

		$! = -7;
		$result = { code => $_parse{code}, diagnostic => $@ };
	}
	else {
		# make sure we have no reported errors on return
		$! = 0;

		# internal only setting
		delete $_parse{reqP};

		foreach my $k (keys %_parse) {
			$result->{$k} = $_parse{$k};
		}
	}

	# clean up parse state tracker
	undef %_parse;

	return $result;
}


sub greeting_parse {
	my $self = shift;

	my $data = shift;

	my $result;

	# parser state info
	$_parse{code}     = 0;
	$_parse{features} = ();
	$_parse{localize} = ("i-default");
	$_parse{uris}     = ();

	# XML parser instance
	my $parser;
	eval {
		$parser  = new XML::Parser::Expat();
		$parser->setHandlers('Start' => \&_greeting_element_start,
		                     'End'   => \&_greeting_element_end,
		                     'Char'  => \&_greeting_cdata);

		$data = xml_normalize($data);
		$parser->parse($data);

		# release internal parser reference(s) -- REQUIRED
		$parser->release();
	};

	if ($@) {
		$self->{log}->entry("notify", "mixer::greeting_parse", $@);

		$! = -7;
		$result = { code => $_parse{code}, diagnostic => $@ };
	}
	else {
		# make sure we have no reported errors on return
		$! = 0;

		$result = { features => $_parse{features},
		            localize => $_parse{localize},
		            uris => $_parse{uris} };
	}

	# clean up parse state tracker
	undef %_parse;

	return $result;
}


#
# framing protocol
#

sub reqrsp {
	my $self = shift;

	my $channel = shift;
	my $data    = shift;

	my $serial = $self->{frame}->REQ($channel, $data);

	if ($self->{faststart}) {
		$self->{faststart} = 0;

		$self->bxxp_greetingaux();
		$self->sep_startaux();
	}

	return $self->rsp($serial);
}


sub rsp {
	my $self = shift;

	my $serial = shift;
	my $eofP = shift || 0;

	my ($data, $frame, $response, $status);

	while (1) {
		$frame = $self->{frame}->read();

		if ($frame && (defined($frame->{keyword}) && $frame->{keyword} ne "RSP")) {
			if (defined($frame->{channel}) && $frame->{channel} == 0 && ! $frame->{data}) {
				$status = ($eofP) ? "+" : "-";

				$self->{frame}->RSP($frame->{serial}, $status);

				if ($eofP) {
					$self->{eofP} = 1;
					return 0;
				}
			}
			elsif (defined($frame->{channel}) && $frame->{channel} == 0) {
				my $err = "<error code='550' xml:lang='en-us'>\n";
				$err .= xml_cdata("don't get me started") . "\n";
				$err .= "</error>";

				$self->{frame}->RSP($frame->{serial}, "-", $err);
			}
			else {
				$response = $self->sep_request($frame);
				if ($! == 0) {
					$data  = "<response reqno='$response->{reqno}'>\n";
					$data .= "<answers />\n";
					$data .= "</response>";

					$status = "+";
				}
				elsif ($! < 0) {
					# internal error
					$data  = "<error code='$response->{code}' xml:lang='en-us'>\n";
					$data .= xml_cdata($response->{diagnostic}) . "\n";
					$data .= "</error>";

					if ($response->{reqno}) {
						my $err = $data;
						$data  = "<response reqno='$response->{reqno}'>\n";
						$data .= $err;
						$data .= "</response>";
					}

					$status = "-";
				}
				else {
					# system error
					$data  = "<error code='451' xml:lang='en-us'>\n";
					$data .= xml_cdata($!) . "\n";
					$data .= "</error>";

					$status = "-";
				}

				$self->{frame}->RSP($frame->{serial}, $status, $data);
			}

			next;
		}

		if ($frame && ($frame->{serial} ne $serial)) {
			# eventually invoke callback for async operations
			;
		}

		return $frame;
	}

	# NOT-REACHED
	0;
}


sub start_error {
	my $p = shift;
	my $attrs = shift;

	$_parse{code} = $attrs->{code};
	if (! $_parse{code}) {
		$p->xpcroak("code attribute missing in <error> element");
	}

	if ($_parse{code} !~ /^\d+$/) {
		$p->xpcroak("code attribute poorly formed in <error> element");
	}

	if ( ($_parse{code} < 100) || ($_parse{code} > 599) ){
		$p->xpcroak("code attribute in <error> element must be three-digit reply code");
	}

	0;
}



#
# library version info
#
sub version {
	my $self = shift;

	return $self->{_VERSION} . "r" . $self->{_RELEASE};
}


#
# debug support
#
use Data::Dumper;

sub _dump {
	my $self = shift;

	print STDERR Dumper($self), "\n";

	0;
}


#
# non-method private subs
#

#
# we want to recognize four different input sets:
#
# <error> ... </error>
#
# <response>
#     <error> ... </error>
# </response>
#
# <response>
#     <answers> ... </answers>
#     <additional> ... </additional>
# </response>
# 
# <request>
#     <notify>
#         <answers> ... </answers>
#         <additional> ... </additional>
#         <deletions> ... </deletions>
#     </notify>
# </request>
#
# state variable        #1      #2      #3      #4
# --------------        --      --      --      --
# rspno                         xx      xx      xx
# envelope                      xx      xx      xx
# answers                               xx      xx
# actualNum                             xx
# reqStamp                              xx      xx
# additional                            xx      xx
# deletions                                     xx
# code                  xx      xx
# diagnostic            xx      xx
# which                 error   error   a/a     a/a/d
# 

#
# handle XML parser start element callbacks
#
sub _element_start {
	my ($p, $name, %attrs) = @_;

	# get context depth (0 == root element)
	my $depth = $p->depth();

	if (($depth - $_parse{reqP}) == -1) {
		$_parse{which} = $name;
		if ($name ne "request") {
			$p->xpcroak("not expecting <$name> element");
		}

		$_parse{rspno} = $attrs{reqno};
		if (! $_parse{rspno}) {
			$p->xpcroak("reqno attribute missing in <request> element");
		}
	}
	elsif (($depth - $_parse{reqP}) == 0) {
		$_parse{which} = $name;

		if ("$name/$depth" eq "response/0") {
			$_parse{envelope} = { %attrs };

			$_parse{rspno} = $attrs{reqno};
			if (! $_parse{rspno}) {
				$p->xpcroak("reqno attribute missing in <response> element");
			}
		}
		elsif ("$name/$depth" eq "error/0") {
			start_error($p, \%attrs);
			$_parse{data} = "";
		}
		elsif ("$name/$depth" eq "notify/1") {
			$_parse{envelope} = { %attrs };
		}
		else {
			$p->xpcroak("not expecting <$name> element");
		}
	}
	elsif (($depth - $_parse{reqP}) == 1) {
		$_parse{which} = $name;

		if ("$name/$depth" eq "answers/1") {
			# set defaults if they don't have value
			$attrs{actualNum} = 0
				unless (exists $attrs{actualNum});
			$attrs{reqStamp} = ""
				unless (exists $attrs{reqStamp});

			$_parse{actualNum} = $attrs{actualNum};
			$_parse{reqStamp} = $attrs{reqStamp};
		}
		elsif ("$name/$depth" eq "answers/2") {
			# set defaults if they don't have value
			$attrs{reqStamp} = ""
				unless (exists $attrs{reqStamp});

			$_parse{reqStamp} = $attrs{reqStamp};
		}
		elsif ("$name/$depth" eq "additional/1") {
			start_error($p, \%attrs);
		}
		elsif ("$name/$depth" eq "additional/2") {
			start_error($p, \%attrs);
		}
		elsif ("$name/$depth" eq "deletions/2") {
			start_error($p, \%attrs);
		}
		elsif ("$name/$depth" eq "error/1") {
			start_error($p, \%attrs);
		}
		else {
			$p->xpcroak("not expecting <$name> element");
		}

		$_parse{data} = "";
	}
	else {
		if ($_parse{which} eq "error") {
			$p->xpcroak("not expecting <$name> element");
		}

		$_parse{data} .= "<$name";
		while (my ($k, $v) = each %attrs) {
			$_parse{data} .= " $k=" . xml_escape_av($v);
		}
		$_parse{data} .= ">";
	}

	0;
}


#
# handle XML parser end element callbacks
#
sub _element_end {
	my ($p, $name) = @_;

	# get context depth (0 == root element)
	my $depth = $p->depth();

	if ( (($depth - $_parse{reqP}) == -1) ||
	     (($depth - $_parse{reqP}) == 0)  ||
	     (($depth - $_parse{reqP}) == 1) ) {
		if ( ($_parse{which} eq "error") && ($name eq "error") ) {
			$_parse{diagnostic} = Space::Util::trim($_parse{data});
			delete $_parse{data};
		}
	}
	elsif (($depth - $_parse{reqP}) == 2) {
		$_parse{data} .= "</$name>";

		my $which = $_parse{which};
		push(@{$_parse{$which}}, $_parse{data});
		delete $_parse{data};
	}
	else {
		$_parse{data} .= "</$name>";
	}

	0;
}


#
# handle XML parser character data callbacks
#
sub _cdata {
	my ($p, $txt) = @_;

	my $depth = $p->depth();

	if ( (($depth - $_parse{reqP}) == 0) ||
	     (($depth - $_parse{reqP}) == 1) ||
	     (($depth - $_parse{reqP}) == 2) )
	{

		$txt = trim($txt);
		if (! $txt) {
			return 0;
		}

		if ($_parse{which} ne "error") {
			$p->xpcroak("unexpected character data");
		}

		$_parse{data} .= $txt;
	}
	else {
		$_parse{data} .= xml_cdata($txt);
	}

	0;
}


#
# parse greeting start element
#
sub _greeting_element_start {
	my ($p, $name, %attrs) = @_;

	# get context depth (0 == root element)
	my $depth = $p->depth();

	if ("$name/$depth" eq "greeting/0") {
		if ($attrs{features}) {
			$_parse{features}= $attrs{features};
			unless ((split(/\s+/, $_parse{features}))[0]) {
				$p->xpcroak("features attribute poorly formed in <greeting> element");
			}
		}

		if ($attrs{localize}) {
			$_parse{localize} = $attrs{localize};
			unless ((split(/\s+/, $_parse{localize}))[0]) {
				$p->xpcroak("localize attribute poorly formed in <greeting> element");
			}
		}
	}
	elsif ("$name/$depth" eq "profile/1") {
		unless (exists($attrs{uri})) {
			$p->xpcroak("profile element missing uri attribute");
		}

		$_parse{uris}->{$attrs{uri}} = 1;
	}
	else {
		$p->xpcroak("not expecting <$name>");
	}

	return 0;
}


#
# parse greeting end element
#
sub _greeting_element_end {
	my ($p, $name) = @_;

	# NO-OP
	return 0;
}


#
# parse greeting cdata
#
sub _greeting_cdata {
	my ($p, $txt) = @_;

	# non-empty content?
	if (trim($txt)) {
		my $err = "not expecting character data in <" . $p->current_element() . ">";
		$p->xpcroak($err);
	}

	return 0;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::Mixer -  Defines a client API for the Perl SpaceKit

=head1 SYNOPSIS

use Space::Mixer;

Error values are indicated when $! is -7 and the return value 
of the method is a hash reference containing two elements:

    code: a three-digit reply code, and,
    diagnostic: a textual explanation.

Non-error return values are specified below for each method.

=head1 DESCRIPTION

This module defines the client API for the Perl SpaceKit.

=head1 METHODS

=over 4

=item new();

This is a class method, the contructor for Space::Mixer.

=item init($log, $server, $args);

Initializes an instance of the mixer.  The parameters are:

    log:
        an object returned by a successful call to
        Space::Log::new.
    server:
        the domain name (or ip address) of the host running the
        BXXP service.
    args:
        a hash reference. The keyword/value pairs are:

        -debug boolean:
            indicates that debugging should be enabled

        -mechanism "anonymous":
            indicates that the SASL ANONYMOUS profile should be
            used for authentication

        -trace string:
            contains the trace information used during the
            SASL ANONYMOUS exchange

        -mechanism "otp":
            indicates that the SASL OTP profile should be used
            for authentication

        -username string:
            contains the user name used during the SASL OTP
            exchange

        -passphrase string:
            contains the passphrase used for OTP SASL exchange

        -passback coderef:
            (if no passphrase is supplied) defines an callback used
            to return the six English words used during the OTP SASL
            exchange.

=item fin();

Finalizes an instance of the mixer, uninitializing the object.

=item fetch($name, $top);

Given a name (e.g., "doc.rfc.2629") that was returned from a
previous invocation to Space::Mixer::search, returns the body of
the corresponding block.  If the block isn't in the cache,
Space::Mixer::search is called automatically.  Because of this,
exceptional returns are possible. In this latter case, the optional
second parameter specifies the top-level XML element of the block,
and often speeds performance.

Applications should use Space::Mixer::fetch rather than implementing
any application-specific caching mechanism.

=item search($payload, $opts);

Given a set-union specification, and, optionally, an ordering
specification, return a hash reference of 0 or more names
corresponding to blocks that satisfy those assertions.  The
elements of the hash are:

    reqno:
        an integer-valued request number;
    answers:
        a list of 0 or more names corresponding to blocks that
        satisfy the assertions; and,
    actualNum:
        an integer-value corresponding to the "actualNum" of the
        "answers" element, if any;
    reqStamp:
        an opaque string value if the "answers" element is part of
        a persistent fetch;
    additional:
        a list of 0 or more names corresponding to blocks that are
        similar to the "answers".

Attributes of the fetch element are controlled via the options hash
reference, e.g.,

    $mixer->search( ... { maxNum => 25 });

However, rather than using "notification => true", use "notification
=> upcall", where "upcall" is a code reference that's invoked on
subsequent notify operations.  Its parameter is one argument, a hash
reference, containing these elements:

    reqno:
        an integer-valued request number for this upcall;
    prevno:
        an integer-value corresponding to the "reqno" of the
        original search;
    answers:
        a list of 0 or more names corresponding to both blocks
        that previously satisfied the assertions and have recently
        changed, or newly-created blocks that satisfy the assertions;
    reqStamp:
        an opaque string value;
    additional:
        a list of 0 or more names corresponding to blocks that are
        similarly to the newly-created blocks in the "answers"; and,
    deletions:
        a list of 0 or more names corresponding to blocks that
        previously satisfied the assertions and have recently been
        deleted.

The procedure should return according to the SYNOPSIS for this
package (i.e., if $! is -7 then the return result should be a
hash reference containing "code" and "diagnostic" elements.)

=item store($blockList, $action, $serial);

Given a list of bodies, each corresponding to a block, store them.
The parameters are:

    blockList:
        a list of hash references
    action:
        one of: "create", "write", "update", or "delete"
    serial:
        if present, the default serial number to use.

    Each hash reference in "blockList" contains two elements:

        name:
            the name of the block; and,
        body:
            the body of the block.

=item lock($subtree);

Creates a lock for a naming scope ("subtree") returning a integer-valued
handle for subsequent use with Space::Mixer::release.

=item release($prevno, $commit);

Releases a persistent operation identified by an integer-valued handle
("reqno"). If the persistent operation is a lock, then $commit indicates
whether the intervening store operations should be committed ("1") or not
("0").

=item version();

Returns the current library version.

=back

=head1 GENERIC MODE

If Space::Mixer::init is invoked with an argument of
-generic => "true", then instead of starting the SEP, use:

=over 4

=item profiles()

Returns a list of profiles (each identified by a URI) from
the server's greeting.

=item create($profile, $data)

Attempts to create a channel for the profile.  The parameters are:

    profile: a URI identifying the profile to bind to the channel,
    data: optional initialization data for the channel.

=item exch($channel, $mime_entity)

Sends a request, returns a response.  The parameters are:

    channel: a channel-number returned by a successful call to
             Space::Mixer::create(),
    mime_entity: a MIME part to be sent as the request.

Note that the MIME part is returned for both normal and error
responses.

=item errscan($mime_entity)

Parses an "error" element, returning a hash reference with the
standard error elements for code and diagnostic.  The parameters
are:

    mime_entity: a MIME part returned by Space::Mixer::exch().

=back

$mixer->fin() is used to finalize the connection.

=head1 EXAMPLES

 # logging support
 use Space::Log;

 # client support
 use Space::Mixer;

 # default client options
 my (%options) = (
   -debug      => "none",
   -mechanism  => "none",
   -passphrase => "",
   -passback   => undef,
   -port       => 10288,
   -privacy    => "none",
   -trace      => "",
   -username   => "",
 );

 # get client options as appropriate...


 # server to connect to
 my $server = "sqa.invisible.net";

 # log token
 my $log = Space::Log->new("log/mixer.log", "mymixer");

 # create new mixer
 my $mixer = Space::Mixer->new();

 if ($mixer) {
   my $response = $mixer->init($log, $server, \%options);
   if ($! == 0) {
     # success: perform fetch, search, or store operation, etc
     ;
   }
   elsif ($! < 0) {
     print STDERR "Error: $response->{code} - $response->{diagnostic}");

     undef $mixer;
   }
   else {
     print STDERR "Error: $response->{code} - $response->{diagnostic}");

     undef $mixer;
   }

   # release mixer data
   $mixer->fin();
 }
 else {
   print STDERR "Unable to create mixer!\n";
 }

 # release log token
 $log->fin();



Please see the C<big.pl> and C<bxxpng.pl> client applications
in the examples/client/ directory of the spacekit distribution
for more complete examples.

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

