#
#  file: OTP.pm
#  auth: delmar dale (ddale@uucom.com)
#  desc: partial implementation of RFC 2289 and RFC 2243
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::OTP;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK  = qw(&compute_otp &MD5_compute &SHA1_compute
                 &hex_to_word &word_to_hex &makeseed);


# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/OTP.pm#10 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

use Digest::MD5;
use Digest::SHA1;

# global package private variables
use vars qw(@words %wordhash);

# word list from RFC 2289
use Space::OTPlib;

#Computes the OTP from challenge, passphrase and password_type (hex or word)
sub compute_otp {

   my($challenge, $passphrase, $password_type) = @_;
   my($retval);

   if ($challenge !~ /^otp-(md5|sha1)\s+\d+\s+\S+(\s+\S+)?$/) {
      return(0);
   }

   if ($password_type !~ /^(word|hex)$/) {
      return(0);
   }

   if ($passphrase !~ /^.{10,63}$/) {
      return(0);
   }

   my($alg_ident, $sequence_int, $seed, $ext) = split(/\s+/, $challenge);
   $ext .= "";

   $seed = lc $seed;
   my $key = $seed . $passphrase;

   if (($alg_ident eq "otp-md5") && ($password_type eq "hex")) {
      $retval = MD5_compute($key, $sequence_int);
      if ($ext =~ /ext/) {
         $retval = "hex:" . $retval;
      }
   } elsif ($alg_ident eq "otp-md5" && $password_type eq "word") {
      $retval = hex_to_word(MD5_compute($key, $sequence_int));
      if ($ext =~ /ext/) {
         $retval = "word:" . $retval;
      }
   } elsif ($alg_ident eq "otp-sha1" && $password_type eq "hex") {
      $retval = SHA1_compute($key, $sequence_int);
      if ($ext =~ /ext/) {
         $retval = "hex:" . $retval;
      }
   } elsif ($alg_ident eq "otp-sha1" && $password_type eq "word") {
      $retval = hex_to_word(SHA1_compute($key, $sequence_int));
      if ($ext =~ /ext/) {
         $retval = "word:" . $retval;
      }
   }
   else {
      carp "Unsupported algorithm/ext type: $alg_ident/$password_type\n";
      return "";
   }

   return($retval);
}


#Computes the MD5 hash
#Returns the hex password
sub MD5_compute {

   my($key, $count) = @_;
   my($ctx, $buf, $digest, $i, $j, $k, @hex, @result, @numresult, $retval);

   $buf = $key;

   #On the first loop "$seed$passphrase" wil be sent through the hash algorithm
   #On subsequent loops the 64 bit hash will be sent through.  Not sure if this
   #is correct or not.

   for($i=0; $i<=$count; $i++) {

     $ctx = Digest::MD5->new;
     $ctx->add($buf);

     #Get 32 Character Hex String
     $digest = $ctx->hexdigest;

     #Split the 32 characters (128) bit into substrings of 2 characters
     @hex = unpack "a2" x 16, $digest;

     #Convert the substrings into numbers
     for ($j=0; $j<=$#hex; $j++) {
         $hex[$j] = hex($hex[$j]);
     }

     #fold the 128 bits to 64 bits using the XOR method described in the RFC
     for ($k=0; $k<8; $k++) {
       $numresult[$k] = $hex[$k] ^ $hex[$k+8];
       $result[$k] = sprintf "%lx", ($hex[$k] ^ $hex[$k+8]);
       if ($result[$k] =~ /^[0-9a-z]$/) {
          $result[$k] = "0" . $result[$k];
       } 
     }

     $retval = "";
     #Re assemble the 64 bits and assign to $buf
     #We now have our 64 bit hash
     foreach (@result) {
       $retval .= $_
     }
     $buf = pack("C*", @numresult);
   }
   #When the loop finishes send the hash
   return($retval);

} #end of MD5_compute


#Computes the SHA1 hash
#Returns the hex password
sub SHA1_compute {

   my($key, $count) = @_;
   my($ctx, $buf, $digest, $i, $j, $k, @hex, $retval, @numresult, @result);

   $buf = $key;

   $ctx = Digest::SHA1->new;

   #On the first loop "$seed$passphrase" wil be sent through the hash algorithm
   #On subsequent loops the 64 bit hash will be sent through.  Not sure if this
   #is correct or not.
   for($k=0; $k<=$count; $k++) {

     $ctx->add($buf);

     #Get 32 Character Hex String
     $digest = $ctx->hexdigest;

     #Split the 40 characters (160) bit into substrings of 8 characters
     @hex = unpack "a8" x 5, $digest;

     #Convert the substrings into numbers
     for ($j=0; $j<=$#hex; $j++) {
         $hex[$j] = hex($hex[$j]);
     }

     $hex[0] = $hex[0] ^ $hex[2];
     $hex[1] = $hex[1] ^ $hex[3];
     $hex[0] = $hex[0] ^ $hex[4];

     for ($i=0,$j=0; $j<8; $i++, $j+=4) {
        $numresult[$j] = $hex[$i] & 255;
        $numresult[$j+1] = ($hex[$i] >> 8) & 255;
        $numresult[$j+2] = ($hex[$i] >> 16) & 255;
        $numresult[$j+3] = ($hex[$i] >> 24) & 255;
        $result[$j] = sprintf "%lx", ($hex[$i] & 255);
        $result[$j+1] = sprintf "%lx", (($hex[$i] >> 8) & 255);
        $result[$j+2] = sprintf "%lx", (($hex[$i] >> 16) & 255);
        $result[$j+3] = sprintf "%lx", (($hex[$i] >> 24) & 255);
     }
     $retval = "";

     #Re assemble the 64 bits and assign to $buf
     #We now have our 64 bit hash
     foreach (@result) {
        if ($_ =~ /^[0-9a-z]$/) {
           $_ = "0" . $_;
        }
        $retval .= $_
     }
     $buf = pack("C*", @numresult);

   }
   #When the loop finishes send the hash
   return($retval);

} #end of MD5_compute



#Converts a hex OTP to 6 word format OTP.
#Returns 0 if there is an error
sub hex_to_word {

   my($hex) = @_;
   my($i,$j,$sum,@wordnum, $retval, @result);

   $hex = lc($hex);
   $hex =~ s/\s//g;

   if ($hex !~ /^[0-9a-f]{16}$/) {
      return(0);
   }

   my($hex1, $hex2) = unpack "a8" x 2, $hex;
   $hex1 = hex($hex1);
   $hex2 = hex($hex2);

   my($mask) = 1;
   my($index) = 2;


   for ($i=0; $i<32; $i++) {
      $result[$index] = $hex2 & $mask;
      $hex2 = $hex2 >> 1;
      $index++;
   }

   for ($i=0; $i<32; $i++) {
      $result[$index] = $hex1 & $mask;
      $hex1 = $hex1 >> 1;
      $index++;
   }

   $sum = 0;
   for ($i=2; $i<= $#result; $i+=2) {
      $sum += ($result[$i+1] << 1) + $result[$i];
   }

   $result[0] = $sum & $mask;
   $result[1] = ($sum & 2) >> 1;

   for ($i=$#result,$j=0; $i>=0; $i-=11,$j++) {

     $wordnum[$j] += $result[$i] << 10;
     $wordnum[$j] += $result[$i-1] << 9;
     $wordnum[$j] += $result[$i-2] << 8;
     $wordnum[$j] += $result[$i-3] << 7;
     $wordnum[$j] += $result[$i-4] << 6;
     $wordnum[$j] += $result[$i-5] << 5;
     $wordnum[$j] += $result[$i-6] << 4;
     $wordnum[$j] += $result[$i-7] << 3;
     $wordnum[$j] += $result[$i-8] << 2;
     $wordnum[$j] += $result[$i-9] << 1;
     $wordnum[$j] += $result[$i-10];

   }

   $retval = $words[$wordnum[0]] . " ";
   $retval .= $words[$wordnum[1]] . " ";
   $retval .= $words[$wordnum[2]] . " ";
   $retval .= $words[$wordnum[3]] . " ";
   $retval .= $words[$wordnum[4]] . " ";
   $retval .= $words[$wordnum[5]];

   return($retval);
   
} #end of hex_to_word


#Converts a 6 word OTP to a hex OTP.
sub word_to_hex {

   my($word) = @_;
   my($mask) = 1;
   my(@bits, $bitcount, $i, $j, $k, $retval, @hex);

   $word = uc($word);
   
   my(@wordarr) = split(' ', $word);

   for ($bitcount=0,$i=5; $i>=0; $bitcount+=11,$i--) {

      $bits[0+ $bitcount] = $wordhash{$wordarr[$i]} & $mask;
      $bits[1+ $bitcount] = ($wordhash{$wordarr[$i]} >> 1) & $mask;
      $bits[2+ $bitcount] = ($wordhash{$wordarr[$i]} >> 2) & $mask;
      $bits[3+ $bitcount] = ($wordhash{$wordarr[$i]} >> 3) & $mask;
      $bits[4+ $bitcount] = ($wordhash{$wordarr[$i]} >> 4) & $mask;
      $bits[5+ $bitcount] = ($wordhash{$wordarr[$i]} >> 5) & $mask;
      $bits[6+ $bitcount] = ($wordhash{$wordarr[$i]} >> 6) & $mask;
      $bits[7+ $bitcount] = ($wordhash{$wordarr[$i]} >> 7) & $mask;
      $bits[8+ $bitcount] = ($wordhash{$wordarr[$i]} >> 8) & $mask;
      $bits[9+ $bitcount] = ($wordhash{$wordarr[$i]} >> 9) & $mask;
      $bits[10+ $bitcount] = ($wordhash{$wordarr[$i]} >> 10) & $mask;

   }

   for ($j=15,$k=2; $j>=0; $j--,$k+=4) {
       $hex[$j] += $bits[$k];
       $hex[$j] += ($bits[$k+1] << 1);
       $hex[$j] += ($bits[$k+2] << 2);
       $hex[$j] += ($bits[$k+3] << 3);
   }

   for ($i=0; $i<16; $i++) {
       $retval .= sprintf "%lx", $hex[$i];
   }

   return($retval);

} #end of word_to_hex


# for hostname()
use Sys::Hostname;

#
# generate a seed value for OTP
#
sub makeseed {
	my $seed;

	$seed = lc((hostname() =~ /^(\w+)/, $1));
	$seed .= time() % 100000;

	if ((my $len = length($seed)) > 16) {
		$seed = substr($seed, $len-16);
	}

	$seed;
}


#
# return current version
#
sub version {
	return $VERSION . "r" . $RELEASE;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::OTP - Library for One Time Password support in the Perl SpaceKit

=head1 SYNOPSIS

use Space::OTP;

=head1 DESCRIPTION

OTP is a Perl module which partially implements the One Time
Password system described in RFC's 2243 and 2289.  It provides
functions to calculate OTP's from an OTP challenge and convert OTP's
to and from the word and hex formats.

=head1 FUNCTIONS

=over 4

=item compute_otp($challenge, $passphrase, $password_type)

Computes the OTP from a given OTP challenge, a passphrase and
password type.  The challenge will have the form

 otp-<algorithm identifier> <sequence integer> <seed>

For example, "otp-md5 487 dog2".  Currently, only the md5 and sha1
algorithms are supported.  The passphrase is usually entered by the
user and contains 10-63 characters.  The password type will be either
"hex" or "word".

=item MD5_compute($key, $sequence)

This function returns a hex OTP generated for the given key and
sequence number using the MD5 algorithm.

=item SHA1_compute($key, $sequence)

This function returns a hex OTP generated for the given key and
sequence number using the SHA1 algorithm.

=item hex_to_word($hex)

This function converts a hex OTP into the 6 word format.

=item word_to_hex($word)

This function converts a 6 word OTP into the hex format.

=back

=head1 EXAMPLES

use Space::OTP qw(&compute_otp &word_to_hex);

$OTP = compute_otp("otp-md5 445 alpha1", "This is my passphrase", "word");

$hex = word_to_hex($OTP);

=head1 AUTHOR

Delmar Dale <ddale@uucom.com>

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

