#
#  file: SASL.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: Space Server Simple Authentication and Security Layer (SASL)
#        implementation
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::SASL;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw();

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/SASL.pm#12 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

use Fcntl;
use Data::Dumper;

# MIME Base64 encode/decode support
use MIME::Base64;

# SASL db support -- this should be replaced with a better DBM solution
use SDBM_File;



#
# local libraries, if any
#

# spacekit logging support
use Space::Log;

# spacekit general utility support
use Space::Util qw(exclfile trim);

# spacekit XML support
use Space::XML qw(xml_normalize);


# some useful constants


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;


#
# namespace variables
#
#  account:  current user
#  accountA: serialized-array of user information
#  privacyL: ordered list of privacy weights
#
# state variables
#
#  log:       object for logging package
#  clientA:   serialzed-array of client information
#  mechanism: corresponding SASL mechanism
#  locks:     list of accounts locked
#
# present only during XML parsing
#
#  code:          reply code
#  authenticator: authenticator element
#  authorization: authorization element
#  response:      response element
#  abortP:        an aborted response
#

#
# instantiate a new sasl object
#
sub new {
	my $this = shift;
	my $type = ref($this) || $this;

	my $self = {};
	bless $self, $type;

	$self->{account}  = "";
	$self->{accountA} = ();

	# available privacy levels
	$self->{privacyL} = { none => 0, weak => 1, strong => 2 };

	# major.minor version info
	$self->{_VERSION} = $VERSION;
	# release (patch) level
	$self->{_RELEASE} = $RELEASE;

	# initialise object if arguments passed in
	if (@_) {
		$self->init(@_);
	}

	$self;
}


#
# initialise sasl info
#
sub init {
	my $self = shift;

	$self->{log}       = shift;
	$self->{clientA}   = shift;
	$self->{mechanism} = shift;
	$self->{authdir}   = shift || "/iw/bxxd/auth";
	$self->{locks}     = ();

	my $authdb = $self->authDB("sasl");

	tie (%{$self->{DB}}, "SDBM_File", $authdb, O_RDWR|O_CREAT, 0660);

	$self;
}
	

#
# finalise sasl session
#
sub fin {
	my $self = shift;

	# remove all outstanding locks
	foreach my $name (keys %{$self->{locks}}) {
		delete $self->{DB}->{"lock.$name"};
	}

	$self->{locks} = ();

	# release SASL db access
	untie %{$self->{DB}};

	0;
}



#
# fetch sasl info
#
sub fetch {
	my $self = shift;

	my $name = shift;

	# clear global error indicator
	$! = 0;

	# do we have a valid sasl username?
	if (my $status = $self->valid_name($name)) {
		$! = -7;
		return $status;
	}

	# read auth info
	my %auth;
	if (my $status = $self->_read_auth($name, \%auth)) {
		$! = -7;
		return { code => 550, diagnostic => "requested user does not exist" };
	}

	$self->{accountA}->{$name} = { %auth };

	if ($self->{mechanism}) {
		# does the requested mechanism match what we provide?
		if (exists($auth{$name}{authentication}) &&
		    ($auth{$name}{authentication}->{mechanism} ne $self->{mechanism}) ) {
			$! = -7;
			return { code => 535, diagnostic => "authentication mechanism mismatch" };	
		}

		# do we require encryption? (problem with global access to Server?)
		my $priv = $auth{$name}{authentication}->{privacy};
		if ($priv && ($self->{privacyL}->{$priv} > $self->{privacyL}->{Space::Server::privacy()}) ) {
			$! = -7;
			return { code => 538, diagnostic => "authentication mechanism requires encryption" };
		}
	}

	return { %auth };
}


#
# store sasl data
#
sub store {
	my $self = shift;

	my $name   = shift;
	my $info   = shift;
	my $action = shift;

	# clear global error indicator
	$! = 0;

	# do we have a valid sasl username?
	if (my $status = $self->valid_name($name)) {
		$! = -7;
		return $status;
	}

	# write auth info
	if (my $status = $self->_write_auth($name, $info, $action)) {
		$! = -7;
		return $status;
	}

	$self->{accountA} = $info;
}


#
# lock user sasl info
#
sub lock {
	my $self = shift;

	my $name = shift;

	# clear global error indicator
	$! = 0;

	# do we have a valid sasl username?
	if (my $status = $self->valid_name($name)) {
		$! = -7;
		return $status;
	}

	$self->{DB}->{"lock.$name"}++;
	if ($self->{DB}->{"lock.$name"} > 1) {
		$self->{DB}->{"lock.$name"}--;

		$! = -7;
		return { code => 450, diagnostic => "authorization profile being updated" };
	}

	$self->{locks}->{$name} = 1;

	0;
}


#
# release previous lock on user sasl info
#
sub release {
	my $self = shift;

	my $name = shift;

	# clear global error indicator
	$! = 0;

	unless (exists $self->{locks}->{$name}) {
		$! = -7;
		return { code => 553, diagnostic => "$name not locked" };
	}

	# remove lock
	delete $self->{DB}->{"lock.$name"};

	# remove lock from locks list
	delete $self->{locks}->{$name};

	0;
}


#
# is name a valid sasl username
#
sub valid_name {
	my $self = shift;

	my $name = shift;

	# clear global error indicator
	$! = 0;

	my @parts = split(/\./, $name);
	if (($#parts < 0) || (! grep /^\w+$/, @parts)) {
		$! = -7;
		return { code => 554, diagnostic => "invalid username: $name" };
	}

	0;
}


#
# generate auth database name
#
sub authDB {
	my $self = shift;

	my $name   = shift;

	# auth base directory -- should get this from caller config
	my $authdir = $self->{authdir};

	return "$authdir/$name";
}


#
# sasl login
#
sub login {
	my $self = shift;

	my $name = shift;
	my $opts = shift || ();

	# clear global error indicator
	$! = 0;

	# append options to info
	@{$self->{clientA}}{keys %$opts} = values %$opts;

	$self->{account} = $self->{clientA}->{name} = $name;

	$self->{log}->entry("stats", "sasl::login", "...");

	0;
}


#
# sasl failure
#
sub failure {
	my $self = shift;

	my $name = shift;
	my $opts = shift;

	# clear global error indicator
	$! = 0;

	my $info = $self->{clientA};

	# append options to info
	@{$info->{keys %$opts}} = values %$opts;

	$info->{name} = $name;

	$self->{log}->entry("stats", "sasl::failure", "...");

	0;
}

#
# is requested action allowed?
#
sub action_allowed {
	my $self = shift;

	my $action  = shift;
	my $subtree = shift;

	# clear global error indicator
	$! = 0;

	my $info = $self->{accountA};
	my $auth = $info->{authorization};
	my $acl  = $info->{acl};

	my $allow = { subtree => $subtree, privs => "" };

	# do we have the subtree?
	foreach my $base (keys %{$acl}) {
		if ( !$base ||
		      $base eq $subtree ||
		      $subtree =~ /^$base/
		   )
		{
			$allow = $acl->{$base};
			last;
		}
	}

	# and do we allow the specific action requested?
	unless ($allow->{privs} =~ /\b$action\b/) {
		$self->{log}->entry("stats", "sasl::action_allowed",
		            "name $self->{account}", "action $action", "subtree $subtree");

		$! = -7;
		return { code => 537, diagnostic => "$action not authorized" };
	}

	0;
}


# parse state variables
my %_parse;

#
# parse SASL message
#
sub parse {
	my $self = shift;

	my $data = shift;

	# clear global error indicator
	$! = 0;

	my $result;

	$_parse{blob}   = "";
	$_parse{code}   = 500;
	$_parse{status} = 0;

	# XML parser instance
	my $parser;
	eval {
		$parser  = new XML::Parser::Expat();
		$parser->setHandlers('Start' => \&_element_start,
		                     'End'   => \&_element_end,
		                     'Char'  => \&_cdata);

		$data = xml_normalize($data);
		$parser->parse($data);

		# release internal parser reference(s) -- REQUIRED
		$parser->release();
	};

	if ($@) {
		$self->{log}->entry("notify", "sasl::parse", $@);
		$! = -7;
		$result = { code => $_parse{code}, diagnostic => $@ };
	}
	else {
		# make sure we have no reported errors on return
		$! = 0;

		$result = { authenticator => $_parse{authenticator},
		            authorization => $_parse{authorization},
		            response      => $_parse{response},
		            abortP        => $_parse{abortP}
		          };
	}

	# clean up temporary state variables
	%_parse = ();

	return $result;
}


#
# library version info
#
sub version {
	my $self = shift;

	return $self->{_VERSION} . "r" . $self->{_RELEASE};
}


#
# read authentication info
#
sub _read_auth {
	my $self = shift;

	my $name = shift;
	my $auth = shift;

	my $status = 0;

	# clear global error indicator
	$! = 0;

	# restore the authentication data structure from the auth DB
	if (defined($self->{DB}->{$name})) {
		$auth->{$name} = eval ( $self->{DB}->{$name} );
	}
	else {
		$! = -7;
		$status = -1;
	}

	return $status;
}


#
# write specified authentication data
#
sub _write_auth {
	my $self = shift;

	my $name   = shift;
	my $auth   = shift;
	my $action = shift;

	# clear global error indicator
	$! = 0;

	# does authorization profile already exist?
	if (($action eq "create") && defined($self->{DB}->{$name})) {
		$! = -7;
		return { code => 554, diagnostic => "authorization profile already exists" };
	}
	elsif (($action eq "delete") && !defined($self->{DB}->{$name})) {
		$! = -7;
		return { code => 554, diagnostic => "authorization profile does not exist" };
	}

	# is the user profile locked?
	unless (exists $self->{locks}->{$name}) {
		$! = -7;
		return { code => 450, diagnostic => "$name not locked" };
	}

	if ($action eq "create" || $action eq "update") {
		my $d = Data::Dumper->new([ $auth ]);

		# some Dump options to help with restore
		$d->Indent(0)->Purity(1)->Quotekeys(0)->Terse(1);

		# dump the authentication data structure to the auth DB
		$self->{DB}->{$name} = $d->Dump();
	}
	elsif ($action eq "delete") {
		delete $self->{DB}->{$name};
	}
	else {
		$! = -7;
		return { code => 554, diagnostic => "$action is not a recognised operation" };
	}

	0;
}


#
# non-method private subs
#

#
# handle XML parser start element callbacks
#
sub _element_start {
	my ($p, $name, %attrs) = @_;

	my $depth = $p->depth();

	if ("$name/$depth" eq "blob/0") {
		$attrs{status} = "continue" unless (exists $attrs{status});
		if ($attrs{status} !~ /^(abort|complete|continue)$/) {
			$_parse{code} = 500;
			$p->xpcroak("status attribute poorly formed in <blob> element");
		}

		$_parse{status} = $attrs{status};
	}
	else {
		$p->xpcroak("not expecting <$name> element");
	}

	0;
}


#
# handle XML parser end element callbacks
#
sub _element_end {
	my ($p, $name) = @_;

	if ($name eq "blob") {
		eval { $_parse{blob} = decode_base64($_parse{blob}); };
		if ($@) {
			$p->xpcroak("invalid content for <blob> element");
		}
	}

	0;
}


#
# handle XML parser character data callbacks
#
sub _cdata {
	my ($p, $txt) = @_;

	$_parse{blob} .= trim($txt);

	0;
}


#
# debug support
#
sub _dump {
	my $self = shift;

	print STDERR Dumper($self), "\n";

	0;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::SASL - Library for Simple Authentication and Security Layer (SASL) used by the Perl SpaceKit.

=head1 SYNOPSIS

use Space::SASL;

=head1 DESCRIPTION

Provides methods for Simple Authentication and Security Layer (SASL) used
by a BXXP server.

=head1 METHODS

=over 4

=item new();

Class constructor, returns a new SASL object on success, an undefined value
otherwise.

=item init($log, $clientA, $mechanism);

Initializes an instance of the API, returning a token used in subsequent calls.
The parameters are:

log:
  a token returned by a successful call to Space::Log::new;

clientA:
  the hash reference passed to the calling profile's init routine; and,

mechanism:
  the mechanism associated with the calling SASL profile (e.g., "anonymous", "otp", and so on).

=item fin();

Finalizes an instance of the API, uninitializing the object.

=item fetch($name);

Given a user name (e.g., "mrose") returns the corresponding authorization profile.

An authorization profile is a hash reference containing these elements:

name:
  the user name;

authentication:
  a hash reference containing information for an authorization profile.  The elements are:

  mechanism:
    the authentication mechanism (e.g., "anonymous" or "otp");

  privacy:
    the minimum required privacy service necessary for authentication
    to be attempted (i.e., "none", "weak", or "strong");

  algorithm (for otp):
    the digest algorithm (e.g., "md5" or "sha1");

  key (for otp):
    the last successful response.

  seed (for otp):
    the seed;

  sequence (for otp):
    the sequence number; and,

authorization:
  a hash reference containing information for an exchange profile. The elements are:

  acl:
    a hash reference of subtree/privilege pairs.  Each pair contains:

  subtree:
    a naming scope (e.g., "doc.rfc"); and,

  privs:
    a list of allowed actions, containing any or all of the possible
    values for the "action" parameter of SASL::action_allowed.

Possible exceptions:

  451 authorization database corrupt

  550 username unknown
  535 authentication mechanism mismatch
  538 authentication mechanism requires encryption
  554 invalid username

=item store($name, $info, $createP);

Given a user name store its authorization profile.  The parameters are:

name:
  the user name;

info:
  its new authorization profile in a hash reference; and,

createP:
  if non-zero, then the user name must not already exist in the authorization database.


Possible exceptions:

  450 username not locked

  554 invalid username
  554 authorization profile already exists

=item lock($name);

Creates a lock for a user name ("name"), even if the user name doesn't exist
in the authorization database.

Possible exceptions:

  450 account being updated

  554 invalid username

=item release($name);

Releases a lock for a user name ("name").

Possible exceptions:

  553 username locked
  554 invalid username

=item login($name, $options);

Sets the authorization profile for the server.  The "options" parameter is an
arbitrary hash reference used for auditing purposes.

=item failure($name, $options);

Notes that an attempt to authentication has failed.  The "options" parameter
is an arbitrary hash reference used for auditing purposes.

=item action_allowed($action, $subtree);

Checks to see if an action is allowed.  The parameters are:

action:
  the requested action, (i.e., one of "create", "delete", "fetch", "lock", "update", "write"); and,

subtree:
  a naming scope (e.g., "doc.rfc").

Possible exceptions:

  537 action not authorized

=item version();

Library version info.

=back

=head1 EXAMPLES

See the etc/sasl-adm.pl application in the examples/server area of the
SpaceKit distribution for a more complete example of using the SASL API.

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

