#
#  file: Server.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: Space server implementation. 
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::Server;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw(BXXP_ERROR BXXP_CALLBACK);

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/Server.pm#11 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# generic XML parsing support
use XML::Parser::Expat;


# spacekit BXXP framing protocol support
use Space::BXXP::Frame;

# spacekit logging support
use Space::Log;

# spacekit general utility support
use Space::Util qw(timer_init timer_start timer_log timer_fin trim);

# spacekit XML support
use Space::XML qw(xml_cdata xml_escape_av xml_error xml_normalize);


# some useful constants

#
# general BXXP error/return code values
#

# BXXP internal error indicator
use constant BXXP_ERROR    => -7;

# TLS callback available
use constant BXXP_CALLBACK => -10;


# other global vars
use vars qw($debug $privacy);


# debug mode?
$debug = 1;


# privacy/encryption required?
$privacy = "none";


#
# instantiate a new bxxp object
#
sub new {
	my $this = shift;
	my $type = ref($this) || $this;

	my $self = {};
	bless $self, $type;

	# current privacy level
	$self->{_privacy} = \$privacy;

	# major.minor version info
	$self->{_VERSION} = $VERSION;
	# release (patch) level
	$self->{_RELEASE} = $RELEASE;

	# init options provided?
	if (@_) {
		$self->init(@_);
	}

	$self;
}


#
# namespace variables
#
# _privacy:  current privacy level
#
# state variables
#
# log:       object for logging package
# inputC:    channel for reading
# outputC:   channel for writing
# plaintext: vector of supported profiles for plaintext operation
# encrypted: vector of supported profiles for encrypted operation
# frameT:    BXXP::Frame object
# serials:   nested array of client-side serial-numbers
#  number:   serial-number
#  rspV:     command to execute on receiving response
# channels:  nested array of channels
#  uri:      URI of profile
#  exchV:    command to execute for data exchange
#  finV:     command to execute on channel release
# callback:  for transport security profiles
#
# present only during XML parsing
#
# _parse:    parse state info
#


#
# initialise bxxp server info
#
sub init {
	my $self = shift;

	# timing support...always on for now
	timer_init();
	if (1) {
		timer_start();
	}

	timer_log("start bxxp::init");

	# clear global error indicator
	$! = 0;

	# logging token
	$self->{log} = shift;

	# client information (hashref)
	$self->{clientA} = shift;

	# input channel (filehandle)
	$self->{inputC} = shift;

	# output channel (filehandle)
	$self->{outputC} = shift || "";
	# map output to input if none supplied
	$self->{outputC} = $self->{inputC}
		unless ($self->{outputC});

	# framing support
	$self->{frameT} = new Space::BXXP::Frame();
	if ($self->{frameT}) {
		# initialise the frame session
		eval { $self->{frameT}->init($self->{log}, $self->{inputC},
		                                 $self->{outputC}, 1); };
		if ($@) {
			$self->{log}->entry("error", "frame::init", $@);
			$! = BXXP_ERROR;
			return $@;
		}
	}
	else {
		my $err = $!;
		$self->{log}->entry("error", "frame::new", $err);
		$! = BXXP_ERROR;
		return $err;
	}

	# list of profiles (hash of hashes)
	my $profiles = shift;

	# what profile support is available?
	my $bootV;

	my $code;

	# get list of plaintext profiles
	foreach my $profile (keys %{$profiles}) {
		print STDERR "profile: ($profile)\n"
			if ($debug > 1);

		# clear global error indicator
		$! = 0;

		$bootV = $profiles->{$profile}->{bootV};
		if (ref($bootV) eq "CODE") {

			eval { $code = $bootV->($self->{log}); };

			if ($@) {
				# eval error
				$self->{log}->entry("error",
				            "Profile::" . $profiles->{$profile}->{module} . ": ", $@);
			}
			elsif ($code) {
				# app error
				$self->{log}->entry("user",
				            "$code->{code}: $code->{diagnostic}");
			}
			else {
				# encrypted or plaintext profile?
				if ($profiles->{$profile}->{privProvider}) {
					$self->{encrypted}->{$profile} = $profiles->{$profile};
				}
				else {
					$self->{plaintext}->{$profile} = $profiles->{$profile};
				}
			}
		}
	}

	# we require at least one plaintext profile for boot
	unless (keys %{$self->{plaintext}}) {
		croak "no bootable profiles!";
	}

	# serials array
	$self->{serials} = ();

	# channels array
	$self->{channels} = ();

	# TLS profile callback
	$self->{callback} = "";

	# major.minor version info
	$self->{_VERSION} = $VERSION;
	# release (patch) level
	$self->{_RELEASE} = $RELEASE;

	if ($debug > 2) {
		$self->_dump();
	}

	timer_log("end bxxp::init");

	0;
}
	

#
# finalise frame session
#
sub fin {
	my $self = shift;

	my $status = shift;

	# clear global error indicator
	$! = 0;

	$self->finaux($status);

	$self->{frameT}->fin();

	timer_fin($self->{log});

	$self = {};

	0;
}


#
# finalise frame session - auxiliary support
#
sub finaux {
	my $self = shift;

	my $status = shift;

	# clear global error indicator
	$! = 0;

	my $channels = $self->{channels};

	# channel cleanup
	my ($channel, $finV);
	foreach my $idx (@{$channels}) {
		$channel = $channels->[$idx];
		if ($finV = $channel->{finV}) {
			eval { $finV->($status); };
			if ($@) {
				# eval error
				$self->{log}->entry("error", $finV, $@);
			}
		}
	}

	$self->{channels} = ();
	$self->{serials} = ();

	0;
}


# can't have strict refs for up/downcall handlers
#no strict 'refs';

#
# server processing loop
#
sub run {
	my $self = shift;

	timer_log("start bxxp::run");

	# clear global error indicator
	$! = 0;

	# which profiles are available?
	my $profiles;
	if (${$self->{_privacy}} eq "none") {
		$profiles = $self->{plaintext};
	}
	else {
		$profiles = $self->{encrypted};
	}

	# generate server greeting message
	my $data = "<greeting localize='en-US'>" . "\n";
	foreach my $profile (sort keys %{$profiles}) {
		$data .= qq|<profile uri="$profile" />| . "\n";
	}
	$data .= "</greeting>";

	# generate initial/greeting response
	eval { $self->{frameT}->RSP(0, "+", $data); };
	if ($@) {
		# RSP error?
		my $err = $@;

		eval { $self->fin("-"); };

		$! = BXXP_ERROR;
		return $err;
	}

	$self->{channels}->[0] = { uri   => "",
	                           exchV => sub { return $self->exch(@_); },
	                           finV  => sub { 0; },
	                         };

	$self->{serials}->[0] = { number => 0,
	                          rspV => sub { return $self->greeting_parse(); },
	                        };

	my $code = 0;
	my ($channel, $downcall, $frame, $number);
	my ($response, $result, $rspV, $status);
	while (1) {

		timer_log("begin frame::read");

		eval { $frame = $self->{frameT}->read(10*60*1000); };

		timer_log("end frame::read");

		if ($@) {
			$code = BXXP_ERROR;
			$result = $frame;
			last;
		}
		elsif ($!) {
			$code = $!;
			$result = $frame;
			last;
		}

		if ($frame->{keyword} ne "REQ") {
			$number = $frame->{serial};

			$response = $self->{serials}->[$number];
			if (! $response) {
				$self->{log}->entry("user",
				            "not expecting $frame->{keyword} $number");
			}
			else {
				$rspV = $response->{rspV};
				eval { $result = $rspV->($frame->{status}, $frame->{data}); };
				if ($@) {
					# response error
					$self->{log}->entry("error", $response->{rspV}, $result);
				}
				undef $self->{serials}->[$number];
			}

			next;
		}

		if ($frame->{channel} == 0 && ! $frame->{data}) {
			$code = 0;
			$result = "";
			last;
		}

		my $idx = $frame->{channel};
		unless ($self->{channels}->[$idx]) {
			$code = BXXP_ERROR;
			$result = "unknown channel";
			last;
		}
		$channel = $self->{channels}->[$idx];

		$self->{callback} = "";
		$downcall = $channel->{exchV};

		timer_log("eval exch");

		eval { $data = $downcall->($frame->{data}); };
		if ($@) {
			$self->{log}->entry("error", "downcall exch", "{$data}");

			$data = "<error code='451' xml:lang='en-us'>\n";
			$data .= xml_cdata($data);
			$data .= "</error>";
		}

		timer_log("end exch");

		timer_log("begin frame::rsp");

		if ($code) {
			$status = "-";
		}
		else {
			$status = "+";
		}

		eval { $code = $self->{frameT}->RSP($frame->{serial}, $status, $data); };
		if ($@) {
			$code = BXXP_ERROR;
			last;
		}

		timer_log("end frame::rsp");
		
		if ($self->{callback}) {
			eval { $result = &$self->{callback}->($self->{inputC}); };
			if ($@) {
				$self->{log}->entry("error", $self->{callback}, $result);
			}

			$self->finaux("+");
			return $self->run();
		}
	}

	if ($code) {
		$status = "-";
	}
	else {
		$status = "+";

		# catch, but ignore, potential errors
		eval { $self->{frameT}->RSP($frame->{serial}, "+"); };
	}

	$self->fin($status);

	timer_log("end bxxp::run");

	$! = $code;
	return $result;
}


# parse state variables
my %_parse = ();

#
# server message exchange support
#
sub exch {
	my $self = shift;

	my $data = shift || "";

	# clear global error indicator
	$! = 0;

	# which profiles are available?
	my $profiles;
	if (${$self->{_privacy}} eq "none") {
		$profiles = $self->{plaintext};
	}
	else {
		$profiles = $self->{encrypted};
	}

	my ($channel, $code, $uri, $profile, $reply, $result);

	$_parse{code}     = 500;
	$_parse{data}     = "";
	$_parse{number}   = -1;
	@{$_parse{pairs}} = ();

	# XML parser instance
	my $parser;
	eval { $parser  = new XML::Parser::Expat(); };
	unless ($parser && ! $@) {

		$self->{log}->entry("error", "bxxp::exch",
		            "XML parser creation failed: $@");

		my $err = "<error code='451' xml:lang='en-us'>\n";
		$err .= "unrecoverable internal XML::Parser error\n";
		$err .= "</error>";

		$! = BXXP_ERROR;
		return $err;
	}

	$parser->setHandlers('Start' => \&_element_start,
	                     'End'   => \&_element_end,
	                     'Char'  => \&_cdata);

	$data = xml_normalize($data);
	eval { $parser->parse($data); };
	if ($@) {
		# parse error
		$self->{log}->entry("notify", "bxxp::exch", $@);

		$result = $@;
		$result .= xml_error($self->{log}, "bxxp", $data, $parser->context());

		# error code to return
		$code = BXXP_ERROR;
	}
	else {
		# process the parsed data
		$profile = "";

		# walk thru the list in pairs of values
		for (my $i = 0 ; $i < $#{$_parse{pairs}} ; $i+=2) {
			my $uri = ${$_parse{pairs}}[$i];
			my $data = ${$_parse{pairs}}[$i+1];

			# do we support this profile?
			if ($profiles->{$uri}) {
				$profile = $profiles->{$uri};
				last;
			}
		}

		if ($self->{channels}->[$_parse{number}]) {
			$_parse{code} = 450;
			$result = "channel already in use";
		}
		elsif ($profile) {

			# start up the new channel
			$self->{frameT}->start($_parse{number});

			# configure channel

			my $init;
			eval {  $init = $profile->{initV}->($self->{log}, "", $self->{clientA},
			             sub { return $self->upcall($_parse{number}, @_); },
			             $profile->{uri}); };

			$self->{channels}->[$_parse{number}] = {
			  uri => $profile->{uri},
			  exchV => sub { return $profile->{exchV}->($init, @_); },
			  finV  => sub { return $profile->{finV}->($init, @_); },
			};

			# additional data to process?
			if ($data = trim($data)) {
				$channel = $self->{channels}->[$_parse{number}];

				eval { $data = $channel->{exchV}->($data); };
				if ($@) {
					$self->{log}->entry("error", $channel->{exchV}, $data);

					$data  = "<error code='451' xml:lang='en-us'>\n";
					$data .= xml_cdata($data) . "\n";
					$data .= "</error>";
				}
				else {
					# NO-OP
					;
				}
			}

			$result = $data;
		}
		else {
			$code = BXXP_ERROR;
			# did we find profile data?
			if ($#{$_parse{pairs}} >= 0) {
				$_parse{code} = 550;
				$result = "all requested profiles are unsupported";
			}
			else {
				$result = "no <profile> elements present";
			}
		}
	}

	# release internal parser reference(s) -- REQUIRED
	$parser->release();

	# my reply code
	$reply = $_parse{code};

	# clean up temporary state variables
	%_parse = ();

	if ($code == 0) {
		my $profile = "<profile uri=" . xml_escape_av($uri);
		if ($result) {
			$profile .= ">\n$result\n</profile>";
		}
		else {
			$profile .= " />";
		}

		$! = 0;
		return $profile;
	}
	elsif ($code == BXXP_ERROR) {
		my $err = "<error code='$reply' xml:lang='en-us'>\n";
		$err .= xml_cdata($result);
		$err .= "</error>";

		$! = BXXP_ERROR;
		return $err;
	}
	else {
		$! = $code;
		return $result;
	}
}


sub greeting_parse {
	my $self = shift;

	my $data = shift;

	# return empty if nothing to parse
	return () unless $data;

	my $result;

	# parser state info
	$_parse{code}     = 0;
	$_parse{features} = ();
	$_parse{localize} = ("i-default");
	$_parse{uris}     = ();

	# XML parser instance
	my $parser;
	eval {
		$parser  = new XML::Parser::Expat();
		$parser->setHandlers('Start' => \&_greeting_element_start,
		                     'End'   => \&_greeting_element_end,
		                     'Char'  => \&_greeting_cdata);

		$data = xml_normalize($data);
		$parser->parse($data);

		# release internal parser reference(s) -- REQUIRED
		$parser->release();
	};

	if ($@) {
		$self->{log}->entry("notify", "server::greeting_parse", $@);

		$! = -7;
		$result = { code => $_parse{code}, diagnostic => $@ };
	}
	else {
		# make sure we have no reported errors on return
		$! = 0;

		$result = { features => $_parse{features},
		            localize => $_parse{localize},
		            uris => $_parse{uris} };
	}

	# clean up parse state tracker
	undef %_parse;

	return $result;
}


#
# profile callback support
#
sub upcall {
	my $self = shift;

	# channel
	my $idx  = shift;
	# payload
	my $data = shift;
	# RSP callback function name
	my $rspV = shift;

	# clear global error indicator
	$! = 0;

	my $number = $self->{frameT}->REQ($idx, $data);

	$self->{serials}->[$number] = { number => $number, rspV => $rspV };

	0;
}


#
# level of privacy required
#
sub privacy {
	my $self = shift;

	return ref($self) ? ${$self->{_privacy}} : $privacy;
}


#
# library version info
#
sub version {
	my $self = shift;

	return __PACKAGE__ . " " . $self->{_VERSION} . "r" . $self->{_RELEASE};
}


#
# debug support
#
use Data::Dumper;

sub _dump {
	my $self = shift;

	print STDERR Dumper($self), "\n";

	0;
}


#
# non-method private subs
#

#
# handle XML parser start element callbacks
#
sub _element_start {
	my ($p, $name, %attrs) = @_;

	# get context depth (0 == root element)
	my $depth = $p->depth();

	if ($depth == 0) {
		if ($name ne "start") {
			$_parse{code} = 501;
			$p->xpcroak("not expecting <$name> element");
		}

		if (! ($_parse{number} = $attrs{number})) {
			$_parse{code} = 501;
			$p->xpcroak("number attribute missing in <$name> element");
		}
		elsif ($_parse{number} !~ /^\d+$/) {
			$_parse{code} = 501;
			$p->xpcroak("number attribute poorly formed in <$name> element");
		}
		elsif (!($_parse{number} % 2) || ($_parse{number} < 1)) {
			$_parse{code} = 501;
			$p->xpcroak("number attribute in <$name> element must be positive and odd");
		}
	}
	elsif ($depth == 1) {
		if ($name ne "profile") {
			$_parse{code} = 501;
			$p->xpcroak("not expecting <$name> element");
		}

		unless ($attrs{uri}) {
			$_parse{code} = 501;
			$p->xpcroak("uri attribute missing in <$name> element");
		}

		push(@{$_parse{pairs}}, $attrs{uri});
		$_parse{data} = "";
	}
	else {
		# build up the element
		my $xml = "<$name";

		while (my ($k, $v) = each %attrs) {
			$xml .= " $k=" . xml_escape_av($v);
		}

		# close the start tag
		$xml .= ">";

		$_parse{data} .= $xml;
	}

	0;
}


#
# handle XML parser end element callbacks
#
sub _element_end {
	my ($p, $name) = @_;

	# get context depth (0 == root element)
	my $depth = $p->depth();

	if ($depth == 0) {
		# NO-OP
		;
	}
	elsif ($depth == 1) {
		push(@{$_parse{pairs}}, $_parse{data});
		$_parse{data} = "";
	}
	else {
		$_parse{data} .= "</$name>";
	}

	0;
}


#
# handle XML parser character data callbacks
#
sub _cdata {
	my ($p, $txt) = @_;

	if ($p->depth() < 2) {
		if (! ($txt = Space::Util::trim($txt)) ) {
			# no data
			return;
		}
		$_parse{code} = 501;
		$p->xpcroak("unexpected character data");
	}

	$_parse{data} .= xml_cdata($txt);
}


#
# parse greeting start element
#
sub _greeting_element_start {
	my ($p, $name, %attrs) = @_;

	# get context depth (0 == root element)
	my $depth = $p->depth();

	if ("$name/$depth" eq "greeting/0") {
		if ($attrs{features}) {
			$_parse{features}= $attrs{features};
			unless ((split(/\s+/, $_parse{features}))[0]) {
				$p->xpcroak("features attribute poorly formed in <greeting> element");
			}
		}

		if ($attrs{localize}) {
			$_parse{localize} = $attrs{localize};
			unless ((split(/\s+/, $_parse{localize}))[0]) {
				$p->xpcroak("localize attribute poorly formed in <greeting> element");
			}
		}
	}
	elsif ("$name/$depth" eq "profile/1") {
		unless (exists($attrs{uri})) {
			$p->xpcroak("profile element missing uri attribute");
		}

		$_parse{uris}->{$attrs{uri}} = 1;
	}
	else {
		$p->xpcroak("not expecting <$name>");
	}

	return 0;
}


#
# parse greeting end element
#
sub _greeting_element_end {
	my ($p, $name) = @_;

	# NO-OP
	return 0;
}


#
# parse greeting cdata
#
sub _greeting_cdata {
	my ($p, $txt) = @_;

	# non-empty content?
	if (trim($txt)) {
		my $err = "not expecting character data in <" . $p->current_element() . ">";
		$p->xpcroak($err);
	}

	return 0;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::Server - Blocks eXtensible eXchange Protocol Server API

=head1 SYNOPSIS

use Space::Server;

=head1 DESCRIPTION

Perl implementation of the BXXP server API.

=head1 METHODS

=over 4

=item new();

Class contructor, returns a new Space::Server object on success, an undefined
value otherwise.

=item init($log, $client, $inputHandle, $outputHandle, profileList);

Initialize server info.  The parameters are:

log:
  Log token returned by successful call to Space::Log::new;

client:
  A hash reference containing client info for address, domain (host), and port;

inputHandle:
  Input filehandle;

outputHandle:
  Output filehandle, defaults to inputHandle;

profileList:
  A hash reference containing information about the profiles available from this
server (i.e., the list of <profile> elements appearing in the greeting above).
The key will be the profile name (URI) with the value being a hash reference
containing the following information:

  privProvider:
    Does this profile provide privacy (TLS) on the channels?

  bootV:
    Profile boot procedure callback (code reference).  Used to
    determine if the profile is run-time loadable by the server.

  initV:
    Profile initialization procedure callback (code reference).

  exchV:
    Profile message exchange procedure callback (code reference).

  finV:
    Profile finalization procedure callback (code reference).

=item fin();

Finalizes an instance of the API, uninitializing the server object.

=item run();

Start server processing loop.

=item privacy();

Returns the level of privacy available from the server.  Currently one of "none", "weak", or "strong".

=item version();

Library version info.

=back

=head1 EXAMPLES

Please see the example server implementation in the F<examples/server/bin/bxxd.pl>
area of the SpaceKit for further details.

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

