#
#  file: Util.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: common utility support
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::Util;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw(boolean
                exclfile tmpfile
                timer_init timer_start timer_fin timer_log
                trim trimleft trimright
                version
               );

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/Util.pm#10 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

use Errno qw(:POSIX EEXIST EINVAL);
use Fcntl;
use FileHandle;
# high resolution system timer values -- we want time()
use Time::HiRes;


# spacekit logging support
use Space::Log;


# other global vars
use vars qw($debug);


# debug mode?
$debug = 0;


# benchmark/timing support
my %clicks = ();
my $clickZ = 0;



#
# determine true-/false-ness of boolean value
#
sub boolean {
	my $val = shift;

	my $bool = undef;

	$bool = 0 if ($val =~ /^(0|false|no|off)$/io);
	$bool = 1 if ($val =~ /^(1|true|yes|on)$/io);

	# start clean
	$! = 0;
	unless (defined($bool)) { $! = EINVAL; }

	return $bool;
}


#
# create an exclusively opened file
#
sub exclfile {
	my $file = shift;
	my $stayP = shift || 0;

	my $fh;

	# up to 10 tries before total failure
	for my $i (0..9) {
		# start clean
		$! = 0;

		$fh = new FileHandle $file, O_RDWR|O_CREAT|O_EXCL;
		if (defined($fh)) {
			if ((print $fh $$ % 65535, " $stayP", "\n") && !$stayP) {
				# automatically closes the file
				undef $fh;
			}

			# print error?
			if ($!) {
				carp "exclfile: $!", "\n"
					if ($debug);

				# automatically closes the file
				undef $fh;

				# remove the file
				unlink $file;
			}

			return $fh;
		}

		sleep 1;
	}

	# set global errno to indicate failure reason
	$! = EEXIST;

	# undef indicates error OR requested file close...confusing
	# this should be an invalid filehandle
	return -1;
}


#
# create a temporary file
#
sub tmpfile {
	my $prefix = shift;
	my $tmpD = shift || "";

	my $fh;

	if (! $tmpD) {
		$tmpD = $ENV{TMP} || "/tmp";
	}

	my $file = $tmpD . "/" . $prefix;
	$file .= $$ % 65535;

	# up to 10 tries to open a tmp file before total failure
	for my $i (0..9) {
		$fh = new FileHandle $file.$i, O_WRONLY|O_CREAT|O_EXCL, 0600;
		last if (defined $fh);
	}

	return ($file, $fh);
}


#
# benchmark/timing support
#

#
# initialise timer info
#
sub timer_init {
	# clear log messages
	%clicks = ();

	#
	# should this be 0 for _init and time for _start or fully mimic the Tcl
	# routine?

	# reset stopwatch
	$clickZ = 0;
}


#
# start timer
#
sub timer_start {
	# start stopwatch
	$clickZ = sprintf("%6.6f", Time::HiRes::time());

	0;
}


#
# finish timer
#
sub timer_fin {
	my $log = shift;

	# don't bother if we don't have a valid log token
	return -1 unless (ref($log));

	my $click0 = $clickZ;
	my $firstP = 1;

	my $click;
	foreach $click (sort (keys %clicks)) {
		my $elapsed = "";
		if ($firstP) {
			$elapsed = sprintf("%10s", "");
			$firstP = 0;
		}
		else {
			$elapsed = sprintf("%10s", sprintf("%3.6f", $click - $click0));
		}

		my $stamp = sprintf("%10s", sprintf("%6.6f", $click));

		my $msg;
		foreach $msg (@{$clicks{$click}}) {
			print STDERR "timer_fin: timing $stamp $elapsed $msg", "\n"
				if ($debug > 1);

			$log->entry("timing", "{$stamp}", "{$elapsed}", "{$msg}");
		}

		$click0 = $click;
	}

	0;
}


#
# timing log routine
#
sub timer_log {
	my $msg = shift;

	# log only has meaning if we are timing
	if ($clickZ) {
		my $now = sprintf("%6.6f", Time::HiRes::time());

		# append the message to the list
		push(@{$clicks{$now-$clickZ}}, $msg);
	}

	0;
}


#
# trim leading/trailing whitespace
#
sub trim {
	my $txt = shift;

	$txt =~ s/^\s+|\s+$//gm;

	$txt;
}


sub trimleft {
	my $txt = shift;

	$txt =~ s/^\s+//gm;

	$txt;
}


sub trimright {
	my $txt = shift;

	$txt =~ s/\s+$//gm;

	$txt;
}


#
# library version info
#
sub version {
	return $VERSION . "r" . $RELEASE;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::Util - Provides utility functions for the Perl SpaceKit

=head1 SYNOPSIS

use Space::Util;

=head1 DESCRIPTION

This Library provides common utility functions for use in the 
Perl SpaceKit.

=head1 FUNCTIONS

=over 4

=item boolean($val);

Determines true-/false-ness of a boolean value and returns a standard
representation for that value, 0 for false and 1 for true.

=item exclfile($file, $stayP);

Creates an exclusively opened file, and returns the filehandle.
$stayP instructs exclfile to keep the file after opening.  The
default behaviour is to delete the file so that it only exists
as long as the file handle is kept open.

=item tmpfile($prefix, $tmpD);

Creates a temporary file, and returns the filename and filehandle.

=item timer_init();

Initialize timer info.

=item timer_start();

Starts the timer.

=item timer_fin($log);

Finishes the timer.  Uses the log token from Space::Log::new
to dump the timing data to the appropriate log.

=item timer_log($msg);

Generate a timestamp log entry.

=item trim($txt);

Removes leading/trailing whitespace from the given string and
returns the altered text.

=item trimleft($txt);

Removes leading whitespace from the given string and
returns the altered text.

=item trimright($txt);

Removes trailing whitespace from the given string and
returns the altered text.

=back

=head1 EXAMPLES

use Space::Util qw(boolean tmpfile trim);


# get standard boolean value for $val
my $bool = boolean($val);

# create temporary file using tmpD (defaults to /tmp) and
# file name $prefix
my ($name, $fh) = tmpfile($prefix, $tmpD);

# open /path/to/file exclusively, and keep the file on disk
my $fh = Space::Util::exclfile("/path/to/file", 1);

# remove leading/trailing whitespace (\s) from $txt
$txt = trim($txt);

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

