#
#  file: XML.pm
#  auth: brad burdick (bburdick@invisible.net)
#  desc: common XML support
#

#
# Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
#
# The contents of this file are subject to the Blocks Public License (the 
# "License"); you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at http://www.invisible.net/
#
# Software distributed under the License is distributed on an "AS IS" basis, 
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
# the specific language governing rights and limitations under the License.
#

# who am i?
package Space::XML;

# version of perl we expect
require 5.004;

use strict;
use Carp;

use Exporter;
use vars qw (@ISA @EXPORT_OK $VERSION);

@ISA       = qw(Exporter);
@EXPORT_OK = qw(xml_cdata
                xml_error xml_escape xml_escape_av
                xml_normalize
                xml_parse xml_pcdata
                xml_unescape);

# current module version
my $Id =<<'EoI';
# $Id: //depot/isms/SpaceKit/blocks.perl/lib/Space/XML.pm#9 $
EoI
my $RELEASE = sprintf("%d", $Id =~ /^# \$Id: .*#(\d+)/);

$VERSION = "1.0";


# additional modules we use

# xml parser support (low-level)
use XML::Parser::Expat;

use Space::Log;
use Space::Util qw(tmpfile trim);


# other global vars
use vars qw($debug %state);


# debug level
$debug = 0;


# used by xml parser to track state info
%state = ();


#
# cdata encode
#
sub xml_cdata {
	my $xml = shift;

	# &amp; must be first to avoid double encoding
	$xml =~ s/\&/\&amp;/gm;
	$xml =~ s/</\&lt;/gm;
	$xml =~ s/\]\]>/\]\]\&gt;/gm;

	return $xml;
}


#
# log XML error
#
sub xml_error {
	my $log    = shift;
	my $prefix = shift;
	my $data   = shift;
	my $stack  = shift || undef;

	my $text = "";

	# always ON for now
	if (1 || $debug > 1) {
 		my ($tmp, $fh) = tmpfile($prefix);
 
		if (defined($fh)) {
			print $fh $data;
			$fh->close();
		}
		else {
			$log->entry("info", "examine $tmp");
		}
	}
		
	if ($#{$stack} >= 0) {
		$text = "\n\nContext:";
		foreach my $frame (@{$stack}) {
			$text .= " <" . $frame->[0] . ">";
		}
	}

	return $text;
}


#
# encode/escape reserved characters
#
sub xml_escape {
	my $xml = shift;

	# avoid double encoding &
	$xml =~ s/\&amp;/\&/gm;

	$xml = XML::Parser::Expat::xml_escape($xml);

	return $xml;
}


#
# encode/escape attribute values properly
#
sub xml_escape_av {
	my $xml = shift;

	# escape all std characters  (& < ' " >)
	my $allP = shift || 1;

	# remove trailing and leading spaces
	$xml = trim($xml);
	
	# avoid double encoding &
	$xml =~ s/\&amp;/\&/gm;

	if ($allP) {
		# avoid double encoding
		$xml =~ s/\&amp;/\&/gm;
		# escape a few additional characters within attribute values
		$xml =~ s/\&/\&amp;/gm;
		$xml =~ s/\</\&lt;/gm;
	}

	$xml =~ s/'/\&apos;/gm;
	$xml =~ s/"/\&quot;/gm;
	$xml =~ s/>/\&gt;/gm;

	return "\"$xml\"";
}


#
# normalize XML data
#
# NOTE:  not sure this is needed.  copied due to reported problem
#        with <![CDATA[...]]> content in TclXML.
#
sub xml_normalize {
	my $text = shift || "";

	my $litS = "<![CDATA[";
	my $litN = length($litS);
	my $litT = "]]>";
	my $litO = length($litT);

	my $xml = "";

	my $pos = -1;
	while (($pos = index($text, $litS)) > -1) {
		my $match = substr($text, 0, $pos);
		$xml .= $match;

		$text = substr($text, $pos+$litN);
		unless (($pos = index($text, $litT)) > -1) {
			# error, missing close to CDATA
			last;
		}

		$match = substr($text, 0, $pos);
		$match =~ s/\&/\&amp;/g;
		$match =~ s/</\&lt;/g;

		$xml .= $match;

		$text = substr($text, $pos+$litO);
	}

	$xml .= $text;

	return $xml;
}


#
# parse XML stream
#
sub xml_parse {
	my $text = shift;

	my $serial = shift || "";

	# result of parse
	my $result;
	my $result_code = 0;

	# global state info
	$state{body}   = "";
	$state{code}   = 500;
	$state{name}   = "";
	$state{serial} = $serial;

	# get parser instance
	my $parser;
	eval { $parser = new XML::Parser::Expat(); };
	unless ($parser and ! $@) {
		# parse error
		$result->{code}       = $state{code};
		$result->{diagnostic} = $@;

		# set global error/return code
		$! = -7;

		return $result;
	}

	$parser->setHandlers('Start' => \&_element_start,
	                     'End'   => \&_element_end,
	                     'Char'  => \&_cdata);

	# trap potential parse errors
	eval { $parser->parse(xml_normalize($text), ErrorContext => 3); };
	if ($@) {
		# parse error
		$! = -7;
		$result->{code}       = $state{code};
		$result->{diagnostic} = $@;
	}
	else {
		$result->{name} = $state{name};
		$result->{body} = trim($state{body}) . "\n";
	}

	# release expat parser internal references 
	$parser->release();

	# reset state info
	%state = ();

	return $result;
}


#
# restore (decode) to parsed character data
#
sub xml_pcdata {
	my $xml = shift;

	my $flattenP = shift || 1;

	if ($flattenP) {
		$xml =~ s/\r/\n/gm;
		$xml =~ s/\n[\s\n]*/\n/gm;
		$xml =~ s/\s*\n\s*/ /gm;
	}

	$xml =~ s/\&lt;/</gm;
	$xml =~ s/\&gt;/>/gm;
	$xml =~ s/\&apos;/'/gm;
	$xml =~ s/\&quot;/"/gm;
	$xml =~ s/\&amp;/\&/gm;

	return $xml;
}


#
# decode/unescape reserved characters
#
sub xml_unescape {
	my($xml) = shift;

	# why no XML::Parser::Expat::xml_unescape?
	$xml =~ s/\&amp;/\&/gm;
	$xml =~ s/\&lt;/</gm;
	$xml =~ s/\&gt;/>/gm;

	return $xml;
}


#
# start element handler for XML parser
#
sub _element_start {
	my($p, $name, %attrs) = @_;

	# is this the root element?
	if ($p->depth() == 0) {
		# get block element name
		$state{name} = $attrs{name};
		unless ($attrs{name}) {
			$state{code} = 501;
			$p->xpcroak("name attribute missing in <$name> element");
		}

		# override block serial number...possibly
		unless ($attrs{serial}) {
			if ($state{serial}) {
				$attrs{serial} = $state{serial};
			}
			else {
				$state{code} = 501;
				$p->xpcroak("serial attribute missing in <$name> element");
			}
		}
	}

	# build up the element
	my $xml = "<$name";

	while (my ($k, $v) = each %attrs) {
		$xml .= " $k=" . xml_escape_av($v);
	}

	# close the start tag
	$xml .= ">";

	$state{body} .= $xml;

	0;
}


#
# end element handler for XML parser
#
sub _element_end {
	my $p    = shift;
	my $name = shift;

	$state{body} .= "</$name>";

	0;
}


#
# character data (CDATA) handler for XML parser
#
sub _cdata {
	my $p     = shift;
	my $cdata = shift;

	$cdata =~ s/\r?\n/\n/g;
	$cdata =~ s/\r/\n/g;

	$state{body} .= xml_cdata($cdata);

	0;
}


#
# library version info
#
sub version {
	return $VERSION . "r" . $RELEASE;
}


# keep require happy -- DO NOT REMOVE
1;

__END__

=head1 NAME

Space::XML - XML library support for the Perl SpaceKit

=head1 SYNOPSIS

use Space::XML;

=head1 DESCRIPTION

This library provides common XML functions for use with the
Perl SpaceKit.

=head1 FUNCTIONS

=over 3

=item xml_cdata($text);

Takes a string and encodes it according to XML's rules for
string values.  Returns the encoded string value.

=item xml_error($log, $prefix, $data, $stack);

Logs XML errors for further inspection.

=item xml_escape($text);

Takes a string argument and encodes it according to XML's
rules for reserved characters (&, <, etc).  Returns the
escaped value.

=item xml_escape_av($text);

Takes a string argument and encodes and quotes it according to
XML's rules for attribute values.  Returns the escaped value.

=item xml_normalize($text);

Takes XML input and expands any CDATA blocks contained therein.
Returns the normalized value.

=item xml_parse($text);

Given the body of a block, contruct its canonical form and return
a hash reference containing two elements:

    name: the name of the block, and
    body: the body of the block.

If an exception occurs, $! will be set to -7 and the return value
will be a hash reference containing two elements:

    code: a three-digit reply code, and,
    diagnostic: a textual explanation.

Possible exceptions:

    500 XML parse error
    501 name attribute missing in top-level element
    501 serial attribute missing in top-level element

=item xml_pcdata($text, $flatten);

The inverse of xml_cdata.  If $flatten is non-zeo, then leading
and trailing whitespace is compressed.  Returns the unencoded value.

=item xml_unescape($text);

Unescape reserved characters.  Returns the unescaped value.

=back

=head1 EXAMPLES

 use Space::XML qw(xml_cdata xml_normalize xml_unescape);

 $txt = xml_cdata($txt);
 $norm = xml_normalize($text);
 $xml = xml_unescape($xml);

=head1 AUTHOR

Brad Burdick <bburdick@invisible.net>.

Copyright (c) 1998-2000, Invisible Worlds, Inc.  All Rights Reserved.

Please see the LICENSE file distributed with the Blocks SpaceKit for
Perl for full details.

=cut

